# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

import logging as pylogging

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
from pathlib import Path

import requests
from sphinx.util import logging

sys.path.insert(0, os.path.abspath(".."))
from odc.stac import __version__ as _odc_stac_version
from scripts import notebook_hash

# isort: off
# extra imports to check env
import odc.stac.bench


# Workaround for https://github.com/agronholm/sphinx-autodoc-typehints/issues/123
# When this https://github.com/agronholm/sphinx-autodoc-typehints/pull/153
# gets merged, we can remove this
class FilterForIssue123(pylogging.Filter):
    def filter(self, record: pylogging.LogRecord) -> bool:
        # You probably should make this check more specific by checking
        # that dataclass name is in the message, so that you don't filter out
        # other meaningful warnings
        return not record.getMessage().startswith("Cannot treat a function")


logging.getLogger("sphinx_autodoc_typehints").logger.addFilter(FilterForIssue123())
# End of a workaround


def ensure_notebooks(dst_folder):
    """
    Download pre-rendered notebooks from a tar archive
    """
    dst_folder = Path(dst_folder)
    if dst_folder.exists():
        print(f"Found pre-rendered notebooks in {dst_folder}")
        return True

    dst_folder.mkdir()
    nb_hash, nb_paths = notebook_hash.compute("../notebooks")
    nb_names = [p.rsplit("/", 1)[-1].rsplit(".", 1)[0] + ".ipynb" for p in nb_paths]

    for nb in nb_names:
        url = f"https://{nb_hash[:16]}--odc-stac-docs.netlify.app/notebooks/{nb}"
        print(f"{url} -> notebooks/{nb}")
        rr = requests.get(url, timeout=5)
        if not rr:
            return False
        with open(dst_folder / nb, "wt", encoding="utf") as dst:
            dst.write(rr.text)

    return True


# working directory is docs/
# download pre-rendered notebooks unless folder is already populated
if not ensure_notebooks("notebooks"):
    notebooks_directory = os.path.abspath("../notebooks")
    raise RuntimeException(
        "There is no cached version of these notebooks. "
        "Build the notebooks before building the documentation. "
        f"Notebooks are located in {notebooks_directory}."
    )

# -- Project information -----------------------------------------------------

project = "odc-stac"
copyright = "2021, ODC"
author = "ODC"

version = ".".join(_odc_stac_version.split(".", 2)[:2])
# The full version, including alpha/beta/rc tags
release = _odc_stac_version


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx_autodoc_typehints",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "sphinx.ext.extlinks",
    "sphinx.ext.mathjax",
    "nbsphinx",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build"]

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "friendly"

autosummary_generate = True

extlinks = {
    "issue": ("https://github.com/opendatacube/odc-stac/issues/%s", "issue %s"),
    "pull": ("https://github.com/opendatacube/odc-stac/pulls/%s", "PR %s"),
}

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable/", None),
    "numpy": ("https://docs.scipy.org/doc/numpy/", None),
    "xarray": ("https://xarray.pydata.org/en/stable/", None),
    "datacube": ("https://datacube-core.readthedocs.io/en/latest/", None),
    "odc-geo": ("https://odc-geo.readthedocs.io/en/latest/", None),
    "pystac": ("https://pystac.readthedocs.io/en/latest/", None),
    "rasterio": ("https://rasterio.readthedocs.io/en/latest/", None),
}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

html_theme_options = {
    "collapse_navigation": False,
    "logo_only": True,
}

# html_logo = '_static/logo.svg'
html_last_updated_fmt = "%b %d, %Y"
html_show_sphinx = False


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

html_css_files = ["xr-fixes.css"]



"""Utilities for benchmarking."""

import importlib
import json
import pickle
from copy import copy
from dataclasses import dataclass, field
from time import sleep
from timeit import default_timer as t_now
from typing import Any, Dict, List, Optional, Tuple, TypeAlias, Union

import affine
import distributed
import numpy as np
import pystac.item
import xarray as xr
from dask.utils import format_bytes
from odc.geo import CRS
from odc.geo.geobox import GeoBox
from odc.geo.xr import ODCExtension

import odc.stac

TimeSample: TypeAlias = Tuple[float, float, float]
"""(t0, t_finished_submit, t_finished_compute)"""

# pylint: disable=too-many-instance-attributes,too-many-locals,too-many-arguments
# pylint: disable=import-outside-toplevel,import-error


@dataclass
class BenchmarkContext:
    """
    Benchmark Context Metadata.

    Normalized representation of the task being benchmarked and
    the environment it is benchmarked in.
    """

    #################################
    # Cluster stats
    cluster_info: Dict[str, Any] = field(repr=False, init=True, compare=False)
    """client.scheduler_info().copy()"""

    nworkers: int = field(init=False)
    """Number of workers in the cluster"""

    nthreads: int = field(init=False)
    """Number of threads in the cluster across all workers"""

    total_ram: int = field(init=False)
    """Total RAM across all workers of the cluster"""

    #################################
    # Data shape stats
    npix: int
    """Total number of output pixels across all bands/timeslices"""

    nbytes: int
    """Total number of output bytes across bands/timeslices"""

    dtype: str
    """Data type of a pixel, assumed to be the same across bands"""

    shape: Tuple[int, int, int, int]
    """Normalized shape in time,band,y,x order"""

    chunks: Tuple[int, int, int, int]
    """Normalized chunk size in time,band,y,x order"""

    #################################
    # Geo
    crs: str
    """Projection used for output"""

    transform: affine.Affine
    """Linear mapping from pixel coordinates to CRS coordinates"""

    #################################
    # Misc
    scenario: str = ""
    """Some human name for this benchmark data"""

    temporal_id: str = ""
    """Time period covered by data in human readable form"""

    method: str = field(compare=False, default="undefined")
    """Some human name for method, (stackstac, odc-stac, rio-xarray)"""

    extras: Dict[str, Any] = field(compare=False, default_factory=dict)
    """Any other parameters to capture"""

    def __post_init__(self) -> None:
        """Extract stats from cluster_info."""
        self.nworkers = len(self.cluster_info["workers"])
        self.nthreads = sum(
            w["nthreads"] for w in self.cluster_info["workers"].values()
        )
        self.total_ram = sum(
            w["memory_limit"] for w in self.cluster_info["workers"].values()
        )

    @property
    def data_signature(self) -> str:
        """Render textual representation of data shape and type."""
        data_dims = ".".join(map(str, self.shape))
        return f"{data_dims}.{self.dtype}"

    @property
    def chunk_signature(self) -> str:
        """Render textual representation of chunk shapes."""
        return ".".join(map(str, self.chunks))

    def render_txt(self, col_width: int = 10) -> str:
        """
        Render textual representation for human consumption.

        :param col_width: Left column width in characters, defaults to 10
        :return: Multiline string representation of self
        """
        nw = col_width

        transorm_txt = f"\n{'':{nw}}".join(str(self.transform).split("\n")[:2])
        transorm_txt = transorm_txt.replace(".00,", ",")
        transorm_txt = transorm_txt.replace(".00|", "|")

        return f"""
{"method":{nw}}: {self.method}
{"Scenario":{nw}}: {self.scenario}
{"T.slice":{nw}}: {self.temporal_id}
{"Data":{nw}}: {self.data_signature},  {format_bytes(self.nbytes)}
{"Chunks":{nw}}: {self.chunk_signature} (T.B.Y.X)
{"GEO":{nw}}: {self.crs}
{"":{nw}}{transorm_txt}
{"Cluster":{nw}}: {self.nworkers} workers, {self.nthreads} threads, {format_bytes(self.total_ram)}
""".strip()

    def render_timing_info(
        self, times: Tuple[float, float, float], col_width: int = 10
    ) -> str:
        """
        Render textual representation of timing result.

        :param times: (t0, t_submit_done, t_done)
        :param col_width: Left column width in characters, defaults to 10
        :return: Multiline string for human consumption
        """
        t0, t1, t2 = times
        t_submit = t1 - t0
        t_elapsed = t2 - t0
        nw = col_width
        return f"""
{"T.Elapsed":{nw}}: {t_elapsed:8.3f} seconds
{"T.Submit":{nw}}: {t_submit:8.3f} seconds
{"Throughput":{nw}}: {self.npix/(t_elapsed*1e+6):8.3f} Mpx/second (overall)
{"":{nw}}| {self.npix/(self.nthreads*t_elapsed*1e+6):8.3f} Mpx/second (per thread)
""".strip()

    @property
    def resolution(self):
        """Extract resolution."""
        sx, _, _, _, sy, *_ = self.transform
        return min(abs(v) for v in [sx, sy])

    def to_pandas_dict(self) -> Dict[str, Any]:
        """Extract parts one would need for analysis of results."""
        return {
            "method": self.method,
            "scenario": self.scenario,
            "data": self.data_signature,
            "chunks": self.chunk_signature,
            "chunks_x": self.chunks[2],
            "chunks_y": self.chunks[3],
            "resolution": self.resolution,
            "crs": self.crs,
            "npix": self.npix,
            "nbytes": self.nbytes,
            "nthreads": self.nthreads,
            "total_ram": self.total_ram,
        }


def collect_context_info(
    client: distributed.Client, xx: Union[xr.DataArray, xr.Dataset], **kw
) -> BenchmarkContext:
    """
    Assemble :class:`~odc.stac.bench.BenchmarkContext` metadata.

    :param client: Dask distributed client
    :param xx: :class:`~xarray.DataArray` or :class:`~xarray.Dataset` that will be benchmarked
    :param kw: Passed on to :class:`~odc.stac.bench.BenchmarkContext` constructor
    :raises ValueError: If ``xx`` is not of the appropriate type
    :return: Populated :class:`~odc.stac.bench.BenchmarkContext` object
    """
    if isinstance(xx, xr.DataArray):
        npix = xx.data.size
        dtype = xx.data.dtype
        nbytes = npix * dtype.itemsize
        _band = getattr(xx, "band", None)
        nb = _band.shape[0] if _band is not None and _band.ndim > 0 else 1
        assert xx.chunks is not None
        _chunks = {k: max(v) for k, v in zip(xx.dims, xx.chunks)}
    elif isinstance(xx, xr.Dataset):
        npix = sum(b.data.size for b in xx.data_vars.values())
        nbytes = sum(b.data.dtype.itemsize * b.data.size for b in xx.data_vars.values())
        dtype, *_ = {b.dtype for b in xx.data_vars.values()}
        nb = len(xx.data_vars)
        sample_band, *_ = xx.data_vars.values()
        assert sample_band.chunks is not None
        _chunks = {k: max(v) for k, v in zip(sample_band.dims, sample_band.chunks)}
    else:
        raise ValueError("Expect one of `xarray.{DataArray,Dataset}` on input")

    assert isinstance(xx.odc, ODCExtension)
    geobox = xx.odc.geobox
    assert geobox is not None
    yx_dims = geobox.dimensions
    ny, nx = (xx[dim].shape[0] for dim in yx_dims)

    time = getattr(xx, "time", None)
    if time is None:
        nt = 1
        temporal_id = "-"
    else:
        time = time.dt.strftime("%Y-%m-%d")
        if time.ndim == 0:
            nt = 1
            temporal_id = time.item()
        else:
            nt = time.shape[0]
            if nt == 1:
                temporal_id = time.data[0]
            else:
                temporal_id = f"{time.data[0]}__{time.data[-1]}"

    ct, cb, cy, cx = (_chunks.get(k, 1) for k in ["time", "band", *yx_dims])
    chunks = (ct, cb, cy, cx)
    geobox = xx.odc.geobox
    if geobox is None or geobox.crs is None:
        raise ValueError("Can't find GEO info")
    assert isinstance(geobox, GeoBox)
    crs = f"epsg:{geobox.crs.epsg}"
    transform = geobox.transform

    return BenchmarkContext(
        client.scheduler_info().copy(),
        npix=npix,
        nbytes=nbytes,
        dtype=str(dtype),
        shape=(nt, nb, ny, nx),
        chunks=chunks,
        crs=crs,
        transform=transform,
        temporal_id=temporal_id,
        **kw,
    )


@dataclass
class BenchLoadParams:
    """Per experiment configuration."""

    scenario: str = ""
    """Name for this scenario"""

    method: str = "odc-stac"
    """Method to use for loading: ``odc-stac|stackstac``"""

    chunks: Tuple[int, int] = (2048, 2048)
    """Chunk size in pixels in ``Y, X`` order"""

    bands: Optional[Tuple[str, ...]] = None
    """Bands to load, defaults to All"""

    resolution: Optional[float] = None
    """Resolution, leave as ``None`` for native"""

    crs: Optional[str] = None
    """Projection to use, leave as ``None`` for native"""

    resampling: Optional[str] = None
    """Set resampling method when reprojecting at load"""

    patch_url: Any = None
    """Accepts ``planetary_computer.sign``"""

    extra: Dict[str, Dict[str, Any]] = field(default_factory=dict)
    """Extra params per ``method``"""

    def with_method(self, method: str) -> "BenchLoadParams":
        """Replace method field only."""
        other = copy(self)
        other.method = method
        return other

    @property
    def epsg(self) -> Optional[int]:
        """Return EPSG code of crs if it was configured, else ``None``."""
        if self.crs is None:
            return None
        return CRS(self.crs).epsg

    @property
    def chunks_as_dict(self) -> Dict[str, float]:
        """Return chunks in dictionary form."""
        return {"y": self.chunks[0], "x": self.chunks[1]}

    def to_json(self, indent=2) -> str:
        """Convert to JSON string."""
        data = copy(self.__dict__)
        if self.patch_url is not None:
            data["patch_url"] = f"{self.patch_url.__module__}.{self.patch_url.__name__}"
        return json.dumps(data, indent=indent)

    @staticmethod
    def from_json(json_text: str) -> "BenchLoadParams":
        """Construct from JSON string."""
        src = json.loads(json_text)

        # convert arrays to tuples
        for k in ("bands", "chunks"):
            v = src.get(k, None)
            if v is not None:
                src[k] = tuple(v)

        cfg = BenchLoadParams(**src)
        if isinstance(cfg.patch_url, str):
            cfg.patch_url = _method_from_string(cfg.patch_url)
        return cfg

    def compute_args(self, method: str = "") -> Dict[str, Any]:
        """Translate into call arguments for a given method."""
        if method == "":
            method = self.method

        extra = {**self.extra.get(method, {})}

        if method == "odc-stac":
            return _trim_dict(
                {
                    "chunks": self.chunks_as_dict,
                    "crs": self.crs,
                    "resolution": self.resolution,
                    "patch_url": self.patch_url,
                    "bands": self.bands,
                    "resampling": self.resampling,
                    **extra,
                }
            )
        if method == "stackstac":
            from rasterio.enums import Resampling

            resampling = None
            if self.resampling is not None:
                resampling = Resampling[self.resampling]

            assets = None
            if self.bands is not None:
                # translate to list, stackstac doesn't like tuple
                assets = list(self.bands)

            extra.setdefault("dtype", "uint16")
            extra.setdefault("fill_value", np.uint16(0))
            extra.setdefault("xy_coords", "center")
            extra.setdefault("rescale", False)

            return _trim_dict(
                {
                    "chunksize": self.chunks[0],
                    "epsg": self.epsg,
                    "resolution": self.resolution,
                    "assets": assets,
                    "resampling": resampling,
                    **extra,
                }
            )

        return {}


def _default_nodata(dtype):
    if dtype.kind == "f":
        return float("nan")
    return dtype.type(0)


def load_from_json(geojson, params: BenchLoadParams, **kw):
    """
    Turn passed in geojson into a Dask array.

    :param geojson: GeoJSON FeatureCollection
    :param params: data loading configuration
    :param kw: passed on to underlying data load function
    """
    # Don't migrate items when loading with stackstac. Migration to stac 1.1
    # removes `proj:epsg` from the item properties that stackstac (==0.5.1 at
    # the time of writing) uses to determine the CRS
    migrate = kw.get("migrate", params.method == "odc-stac")
    all_items = [
        pystac.item.Item.from_dict(f, migrate=migrate) for f in geojson["features"]
    ]

    opts = params.compute_args()
    opts.update(**kw)

    if params.method == "odc-stac":
        xx = odc.stac.load(all_items, **opts)
    elif params.method == "stackstac":
        import stackstac

        patch_url = params.patch_url
        if patch_url is None:
            patch_url = lambda x: x  # pylint: disable=unnecessary-lambda-assignment

        _items = [patch_url(item).to_dict() for item in all_items]
        xx = stackstac.stack(_items, **opts)
        if np.unique(xx.time.data).shape != xx.time.shape:
            nodata = opts.get("fill_value", _default_nodata(xx.dtype))
            xx = xx.groupby("time").map(stackstac.mosaic, nodata=nodata)

        if xx.odc.geobox.transform != xx.spec.transform:
            # work around issue 93 in stackstac
            xx.y.data[:] = xx.y.data - xx.spec.resolutions_xy[1]
    else:
        raise ValueError(f"Unsupported method:'{params.method}'")

    xx.attrs["load_params"] = params
    xx.attrs["_opts"] = opts
    return xx


def run_bench(
    xx: Union[xr.DataArray, xr.Dataset],
    client: distributed.Client,
    ntimes: int = 1,
    col_width: int = 12,
    restart_sleep: float = 0,
    results_file: Optional[str] = None,
) -> Tuple[BenchmarkContext, List[TimeSample]]:
    """
    Run same configuration multiple times and resport timing.

    :param xx: Dask graph to persist to ram
    :param client: Dask client to test on (will be restarted between runs)
    :param ntimes: How many rounds to run (default: 1)
    :param col_width: First column width in characters
    :param restart_sleep: Number of seconds to sleep after ``client.restart()``
    :param results_file: If set pickle results to this file, it will be overwritten
                         after every run.
    :returns: :class:`odc.stac.bench.BenchmarkContext` and timing info per run.

    Reported timing info is a triple of ``(t0, t_finished_submit, t_finished_persist)``
    """
    params = xx.attrs.get("load_params", None)

    extra = {}
    if params is not None:
        extra["method"] = params.method
        extra["scenario"] = params.scenario

    bench_ctx = collect_context_info(client, xx, **extra)
    samples = []
    _xx = None

    try:
        print(bench_ctx.render_txt(col_width))
        for _ in range(ntimes):
            client.restart()
            sleep(restart_sleep)

            t0 = t_now()

            _xx = client.persist(xx)
            t1 = t_now()

            _ = distributed.wait(_xx)
            t2 = t_now()

            times = (t0, t1, t2)
            print("-" * 60)
            print(bench_ctx.render_timing_info(times, col_width))
            samples.append(times)

            if results_file is not None:
                with open(results_file, "wb") as dst:
                    pickle.dump({"context": bench_ctx, "samples": samples}, dst)
    except KeyboardInterrupt:
        print("Aborting early upon request")
        if _xx is not None:
            client.cancel(_xx)

    return bench_ctx, samples


def _trim_dict(d: Dict[str, Any]) -> Dict[str, Any]:
    return {k: v for k, v in d.items() if v is not None}


def _method_from_string(method_path: str) -> Any:
    module_name, method_name = method_path.rsplit(".", 1)
    return getattr(importlib.import_module(module_name), method_name)



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.13.6
#   kernelspec:
#     display_name: 'Python 3.8.12 64-bit (''stac'': conda)'
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Access Sentinel 2 Analysis Ready Data from Digital Earth Africa
#
# [![Binder](https://mybinder.org/badge_logo.svg)](https://mybinder.org/v2/gh/opendatacube/odc-stac/develop?labpath=notebooks%2Fstac-load-S2-deafrica.ipynb)
#
# https://explorer.digitalearth.africa/products/s2_l2a

# %% [markdown]
# ## Import Required Packages

# %%
from pystac_client import Client

from odc.stac import configure_rio, stac_load

# %% [markdown]
# ## Set Collection Configuration
#
# The configuration dictionary is determined from the product's definition, available at https://explorer.digitalearth.africa/products/s2_l2a#definition-doc
#
# All assets except SCL have the same configuration. SCL uses `uint8` rather than `uint16`.
#
# In the configuration, we also supply the aliases for each band. This means we can load data by band name rather than band number.

# %%
config = {
    "s2_l2a": {
        "assets": {
            "*": {
                "data_type": "uint16",
                "nodata": 0,
                "unit": "1",
            },
            "SCL": {
                "data_type": "uint8",
                "nodata": 0,
                "unit": "1",
            },
        },
        "aliases": {
            "costal_aerosol": "B01",
            "blue": "B02",
            "green": "B03",
            "red": "B04",
            "red_edge_1": "B05",
            "red_edge_2": "B06",
            "red_edge_3": "B07",
            "nir": "B08",
            "nir_narrow": "B08A",
            "water_vapour": "B09",
            "swir_1": "B11",
            "swir_2": "B12",
            "mask": "SCL",
            "aerosol_optical_thickness": "AOT",
            "scene_average_water_vapour": "WVP",
        },
    }
}


# %% [markdown]
# ## Set AWS Configuration
#
# Digital Earth Africa data is stored on S3 in Cape Town, Africa. To load the data, we must configure rasterio with the appropriate AWS S3 endpoint. This can be done with the `odc.stac.configure_rio` function. Documentation for this function is available at https://odc-stac.readthedocs.io/en/latest/_api/odc.stac.configure_rio.html#odc.stac.configure_rio.
#
# The configuration below must be used when loading any Digital Earth Africa data through the STAC API.

# %%
configure_rio(
    cloud_defaults=True,
    aws={"aws_unsigned": True},
    AWS_S3_ENDPOINT="s3.af-south-1.amazonaws.com",
)


# %% [markdown]
# ## Connect to the Digital Earth Africa STAC Catalog

# %%
# Open the stac catalogue
catalog = Client.open("https://explorer.digitalearth.africa/stac")


# %% [markdown]
# ## Find STAC Items to Load
#
# ### Define query parameters

# %%
# Set a bounding box
# [xmin, ymin, xmax, ymax] in latitude and longitude
bbox = [37.76, 12.49, 37.77, 12.50]

# Set a start and end date
start_date = "2020-09-01"
end_date = "2020-12-01"

# Set the STAC collections
collections = ["s2_l2a"]


# %% [markdown]
# ### Construct query and get items from catalog

# %%
# Build a query with the set parameters
query = catalog.search(
    bbox=bbox, collections=collections, datetime=f"{start_date}/{end_date}"
)

# Search the STAC catalog for all items matching the query
items = list(query.items())
print(f"Found: {len(items):d} datasets")

# %% [markdown]
# ## Load the Data
#
# In this step, we specify the desired coordinate system, resolution (here 20m), and bands to load. We also pass the bounding box to the `stac_load` function to only load the requested data. Since the band aliases are contained in the `config` dictionary, bands can be loaded using these aliaes (e.g. `"red"` instead of `"B04"` below).
#
# The data will be lazy-loaded with dask, meaning that is won't be loaded into memory until necessary, such as when it is displayed.

# %%
crs = "EPSG:6933"
resolution = 20

ds = stac_load(
    items,
    bands=("red", "green", "blue", "nir"),
    crs=crs,
    resolution=resolution,
    chunks={},
    groupby="solar_day",
    stac_cfg=config,
    bbox=bbox,
)

# View the Xarray Dataset
ds


# %% [markdown]
# ### Compute a band index
#
# After loading the data, you can perform standard Xarray operations, such as calculating and plotting the normalised difference vegetation index (NDVI). The `.compute()` method triggers Dask to load the data into memory, so running this step may take a few minutes.

# %%
ds["NDVI"] = (ds.nir - ds.red) / (ds.nir + ds.red)


ds.NDVI.compute().plot(col="time", col_wrap=6, vmin=0, vmax=1)



"""
Test for SQS to DC tool
"""

import json
from pathlib import Path

import pystac
import pystac.collection
import pystac.item
import pytest
from odc.geo.data import country_geom

TEST_DATA_FOLDER: Path = Path(__file__).parent.joinpath("data")
PARTIAL_PROJ_STAC: str = "only_crs_proj.json"
GA_LANDSAT_STAC: str = "ga_ls8c_ard_3-1-0_088080_2020-05-25_final.stac-item.json"
SENTINEL_STAC_COLLECTION: str = "sentinel-2-l2a.collection.json"
SENTINEL_STAC: str = "S2A_28QCH_20200714_0_L2A.json"
SENTINEL_STAC_MS: str = "S2B_MSIL2A_20190629T212529_R043_T06VVN_20201006T080531.json"
SENTINEL_STAC_MS_RASTER_EXT: str = (
    "S2B_MSIL2A_20190629T212529_R043_T06VVN_20201006T080531_raster_ext.json"
)
USGS_LANDSAT_STAC_v1b: str = "LC08_L2SR_081119_20200101_20200823_02_T2.json"
USGS_LANDSAT_STAC_v1: str = "LC08_L2SP_028030_20200114_20200824_02_T1_SR.json"
USGS_LANDSAT_STAC_v1_1_1: str = "LE07_L2SP_044033_20210329_20210424_02_T1_SR.json"
LIDAR_STAC: str = "lidar_dem.json"
BENCH_SITE1: str = "site1-20200606-tall-strip-africa.geojson"
BENCH_SITE2: str = "site2-2020_jun_jul-35MNM.geojson"

# pylint: disable=redefined-outer-name


@pytest.fixture(scope="session")
def test_data_dir():
    return TEST_DATA_FOLDER


@pytest.fixture
def partial_proj_stac():
    return pystac.item.Item.from_file(str(TEST_DATA_FOLDER.joinpath(PARTIAL_PROJ_STAC)))


@pytest.fixture
def no_bands_stac(partial_proj_stac):
    partial_proj_stac.assets.clear()
    return partial_proj_stac


@pytest.fixture
def usgs_landsat_stac_v1():
    return pystac.item.Item.from_file(
        str(TEST_DATA_FOLDER.joinpath(USGS_LANDSAT_STAC_v1))
    )


@pytest.fixture
def usgs_landsat_stac_v1b():
    return pystac.item.Item.from_file(
        str(TEST_DATA_FOLDER.joinpath(USGS_LANDSAT_STAC_v1b))
    )


@pytest.fixture
def usgs_landsat_stac_v1_1_1():
    return pystac.item.Item.from_file(
        str(TEST_DATA_FOLDER.joinpath(USGS_LANDSAT_STAC_v1_1_1))
    )


@pytest.fixture
def ga_landsat_stac():
    return pystac.item.Item.from_file(str(TEST_DATA_FOLDER.joinpath(GA_LANDSAT_STAC)))


@pytest.fixture
def lidar_stac():
    return pystac.item.Item.from_file(str(TEST_DATA_FOLDER.joinpath(LIDAR_STAC)))


@pytest.fixture
def sentinel_stac():
    return pystac.item.Item.from_file(str(TEST_DATA_FOLDER.joinpath(SENTINEL_STAC)))


@pytest.fixture
def sentinel_stac_ms_json():
    with TEST_DATA_FOLDER.joinpath(SENTINEL_STAC_MS).open("r", encoding="utf") as f:
        return json.load(f)


@pytest.fixture
def bench_site1():
    with TEST_DATA_FOLDER.joinpath(BENCH_SITE1).open("r", encoding="utf") as f:
        return _strip_links(json.load(f))


@pytest.fixture
def bench_site2():
    with TEST_DATA_FOLDER.joinpath(BENCH_SITE2).open("r", encoding="utf") as f:
        return _strip_links(json.load(f))


@pytest.fixture
def sentinel_stac_ms():
    return pystac.item.Item.from_file(str(TEST_DATA_FOLDER.joinpath(SENTINEL_STAC_MS)))


@pytest.fixture
def sentinel_stac_ms_no_ext(sentinel_stac_ms_json):
    metadata = dict(sentinel_stac_ms_json)
    metadata["stac_extensions"] = []
    return pystac.item.Item.from_dict(metadata)


@pytest.fixture
def sentinel_stac_ms_with_raster_ext():
    return pystac.item.Item.from_file(
        str(TEST_DATA_FOLDER.joinpath(SENTINEL_STAC_MS_RASTER_EXT))
    )


@pytest.fixture
def sentinel_stac_collection():
    return pystac.collection.Collection.from_file(
        str(TEST_DATA_FOLDER.joinpath(SENTINEL_STAC_COLLECTION))
    )


@pytest.fixture
def relative_href_only(ga_landsat_stac: pystac.item.Item):
    item = pystac.Item.from_dict(ga_landsat_stac.to_dict())
    item = item.make_asset_hrefs_relative()
    assert isinstance(item, pystac.Item)
    item.remove_links("self")
    return item


@pytest.fixture
def sample_geojson():
    return {
        "type": "FeatureCollection",
        "features": [
            {
                "type": "Feature",
                "properties": {"name": "Kangaroo Island"},
                "geometry": {
                    "type": "Polygon",
                    "coordinates": [
                        [
                            [136.351318359375, -35.78217070326606],
                            [136.7303466796875, -36.16448788632062],
                            [137.5323486328125, -36.16005298551352],
                            [137.8179931640625, -35.933540642493114],
                            [138.0816650390625, -36.05798104702501],
                            [138.2025146484375, -35.74205383068035],
                            [137.5653076171875, -35.46066995149529],
                            [136.351318359375, -35.78217070326606],
                        ]
                    ],
                },
            }
        ],
    }


def _strip_links(gjson):
    for item in gjson["features"]:
        item["links"] = []
    return gjson


@pytest.fixture()
def gpd_iso3():
    def _get(iso3: str, crs=None):
        return country_geom(iso3.upper(), crs=crs)

    yield _get



# pylint: disable=redefined-outer-name,missing-module-docstring,missing-function-docstring,missing-class-docstring
# pylint: disable=use-implicit-booleaness-not-comparison,protected-access
from __future__ import annotations

import numpy as np
import pytest
import xarray as xr
from odc.loader.types import AuxBandMetadata, AuxDataSource, AuxLoadParams

from odc.stac._mdtools import StacAuxReader
from odc.stac.model import PropertyLoadRequest, RasterBandMetadata
from odc.stac.testing.stac import b_, mk_parsed_item


class TestStacAuxReader:
    """Test cases for StacAuxReader class."""

    def test_init(self):
        """Test StacAuxReader initialization."""
        reader = StacAuxReader()
        assert reader is not None

    def test_read_basic(self):
        """Test basic reading functionality with simple data."""
        reader = StacAuxReader()

        # Create test data sources
        prop_req = PropertyLoadRequest(key="test_prop", dtype="float32", units="1")
        aux_source = AuxDataSource(
            uri="virtual://test/1",
            subdataset=None,
            meta=AuxBandMetadata("float32", None, "1", driver_data=prop_req),
            driver_data=42.0,
        )

        srcs = [[aux_source]]  # Single time step, single source
        cfg = AuxLoadParams(
            dtype="float32",
            fill_value=None,
            meta=AuxBandMetadata("float32", None, "1", driver_data=prop_req),
        )
        used_names = {"time"}
        time_coord = xr.DataArray(
            [np.datetime64("2020-01-01")], dims=["time"], name="time"
        )
        available_coords = {"time": time_coord}
        ctx = None

        result = reader.read(srcs, cfg, used_names, available_coords, ctx)

        assert isinstance(result, xr.DataArray)
        assert result.dims == ("time",)
        assert result.shape == (1,)
        assert result.values[0] == 42.0
        assert result.attrs["units"] == "1"

    def test_read_empty_sources(self):
        """Test reading with empty sources (should use fill value)."""
        reader = StacAuxReader()

        prop_req = PropertyLoadRequest(
            key="test_prop", dtype="float32", units="1", nodata=-999
        )

        srcs = [[]]  # Empty sources
        cfg = AuxLoadParams(
            dtype="float32",
            fill_value=-999,
            meta=AuxBandMetadata("float32", -999, "1", driver_data=prop_req),
        )
        used_names = {"time"}
        time_coord = xr.DataArray(
            [np.datetime64("2020-01-01")], dims=["time"], name="time"
        )
        available_coords = {"time": time_coord}
        ctx = None

        result = reader.read(srcs, cfg, used_names, available_coords, ctx)

        assert isinstance(result, xr.DataArray)
        assert result.dims == ("time",)
        assert result.shape == (1,)
        assert result.values[0] == -999
        assert result.attrs["units"] == "1"
        assert result.attrs["nodata"] == -999

    def test_read_multiple_sources(self):
        """Test reading with multiple sources in a single time step."""
        reader = StacAuxReader()

        prop_req = PropertyLoadRequest(key="test_prop", dtype="float32", units="1")

        # Multiple sources for same time step
        aux_source1 = AuxDataSource(
            uri="virtual://test/1",
            subdataset=None,
            meta=AuxBandMetadata("float32", None, "1", driver_data=prop_req),
            driver_data=10.0,
        )
        aux_source2 = AuxDataSource(
            uri="virtual://test/2",
            subdataset=None,
            meta=AuxBandMetadata("float32", None, "1", driver_data=prop_req),
            driver_data=20.0,
        )

        srcs = [[aux_source1, aux_source2]]  # Single time step, multiple sources
        cfg = AuxLoadParams(
            dtype="float32",
            fill_value=None,
            meta=AuxBandMetadata("float32", None, "1", driver_data=prop_req),
        )
        used_names = {"time"}
        time_coord = xr.DataArray(
            [np.datetime64("2020-01-01")], dims=["time"], name="time"
        )
        available_coords = {"time": time_coord}
        ctx = None

        result = reader.read(srcs, cfg, used_names, available_coords, ctx)

        assert isinstance(result, xr.DataArray)
        assert result.dims == ("time",)
        assert result.shape == (1,)
        # Should use the fuser function to combine values
        assert result.values[0] == 15.0  # Default fuser averages values (10 + 20) / 2

    def test_read_multiple_time_steps(self):
        """Test reading with multiple time steps."""
        reader = StacAuxReader()

        prop_req = PropertyLoadRequest(key="test_prop", dtype="float32", units="1")

        aux_source1 = AuxDataSource(
            uri="virtual://test/1",
            subdataset=None,
            meta=AuxBandMetadata("float32", None, "1", driver_data=prop_req),
            driver_data=10.0,
        )
        aux_source2 = AuxDataSource(
            uri="virtual://test/2",
            subdataset=None,
            meta=AuxBandMetadata("float32", None, "1", driver_data=prop_req),
            driver_data=20.0,
        )

        srcs = [[aux_source1], [aux_source2]]  # Two time steps
        cfg = AuxLoadParams(
            dtype="float32",
            fill_value=None,
            meta=AuxBandMetadata("float32", None, "1", driver_data=prop_req),
        )
        used_names = {"time"}
        time_coord = xr.DataArray(
            [np.datetime64("2020-01-01"), np.datetime64("2020-01-02")],
            dims=["time"],
            name="time",
        )
        available_coords = {"time": time_coord}
        ctx = None

        result = reader.read(srcs, cfg, used_names, available_coords, ctx)

        assert isinstance(result, xr.DataArray)
        assert result.dims == ("time",)
        assert result.shape == (2,)
        assert result.values[0] == 10.0
        assert result.values[1] == 20.0

    def test_read_with_none_values(self):
        """Test reading with None values in driver_data."""
        reader = StacAuxReader()

        prop_req = PropertyLoadRequest(key="test_prop", dtype="float32", units="1")

        aux_source1 = AuxDataSource(
            uri="virtual://test/1",
            subdataset=None,
            meta=AuxBandMetadata("float32", None, "1", driver_data=prop_req),
            driver_data=None,
        )
        aux_source2 = AuxDataSource(
            uri="virtual://test/2",
            subdataset=None,
            meta=AuxBandMetadata("float32", None, "1", driver_data=prop_req),
            driver_data=20.0,
        )

        srcs = [[aux_source1, aux_source2]]
        cfg = AuxLoadParams(
            dtype="float32",
            fill_value=-999,
            meta=AuxBandMetadata("float32", -999, "1", driver_data=prop_req),
        )
        used_names = {"time"}
        time_coord = xr.DataArray(
            [np.datetime64("2020-01-01")], dims=["time"], name="time"
        )
        available_coords = {"time": time_coord}
        ctx = None

        result = reader.read(srcs, cfg, used_names, available_coords, ctx)

        assert isinstance(result, xr.DataArray)
        assert result.dims == ("time",)
        assert result.shape == (1,)
        # Should filter out None values and use only 20.0
        assert result.values[0] == 20.0

    def test_read_different_dtypes(self):
        """Test reading with different data types."""
        reader = StacAuxReader()

        # Test int32
        prop_req_int = PropertyLoadRequest(key="test_prop", dtype="int32", units="1")
        aux_source_int = AuxDataSource(
            uri="virtual://test/1",
            subdataset=None,
            meta=AuxBandMetadata("int32", None, "1", driver_data=prop_req_int),
            driver_data=42,
        )

        srcs = [[aux_source_int]]
        cfg = AuxLoadParams(
            dtype="int32",
            fill_value=None,
            meta=AuxBandMetadata("int32", None, "1", driver_data=prop_req_int),
        )
        used_names = {"time"}
        time_coord = xr.DataArray(
            [np.datetime64("2020-01-01")], dims=["time"], name="time"
        )
        available_coords = {"time": time_coord}
        ctx = None

        result = reader.read(srcs, cfg, used_names, available_coords, ctx)

        assert result.dtype == np.int32
        assert result.values[0] == 42

    def test_read_with_custom_fuser(self):
        """Test reading with custom fuser function."""
        reader = StacAuxReader()

        def custom_fuser(values):
            return sum(values)  # Sum all values

        prop_req = PropertyLoadRequest(
            key="test_prop", dtype="float32", units="1", fuser=custom_fuser
        )

        aux_source1 = AuxDataSource(
            uri="virtual://test/1",
            subdataset=None,
            meta=AuxBandMetadata("float32", None, "1", driver_data=prop_req),
            driver_data=10.0,
        )
        aux_source2 = AuxDataSource(
            uri="virtual://test/2",
            subdataset=None,
            meta=AuxBandMetadata("float32", None, "1", driver_data=prop_req),
            driver_data=20.0,
        )

        srcs = [[aux_source1, aux_source2]]
        cfg = AuxLoadParams(
            dtype="float32",
            fill_value=None,
            meta=AuxBandMetadata("float32", None, "1", driver_data=prop_req),
        )
        used_names = {"time"}
        time_coord = xr.DataArray(
            [np.datetime64("2020-01-01")], dims=["time"], name="time"
        )
        available_coords = {"time": time_coord}
        ctx = None

        result = reader.read(srcs, cfg, used_names, available_coords, ctx)

        assert result.values[0] == 30.0  # 10 + 20

    def test_read_with_nodata_fill_value(self):
        """Test reading with nodata fill value."""
        reader = StacAuxReader()

        prop_req = PropertyLoadRequest(
            key="test_prop", dtype="float32", units="1", nodata=-999
        )

        aux_source = AuxDataSource(
            uri="virtual://test/1",
            subdataset=None,
            meta=AuxBandMetadata("float32", -999, "1", driver_data=prop_req),
            driver_data=42.0,
        )

        srcs = [[aux_source]]
        cfg = AuxLoadParams(
            dtype="float32",
            fill_value=-999,
            meta=AuxBandMetadata("float32", -999, "1", driver_data=prop_req),
        )
        used_names = {"time"}
        time_coord = xr.DataArray(
            [np.datetime64("2020-01-01")], dims=["time"], name="time"
        )
        available_coords = {"time": time_coord}
        ctx = None

        result = reader.read(srcs, cfg, used_names, available_coords, ctx)

        assert result.attrs["nodata"] == -999

    def test_read_with_string_data(self):
        """Test reading with string data."""
        reader = StacAuxReader()

        prop_req = PropertyLoadRequest(key="test_prop", dtype="object", units="1")

        aux_source = AuxDataSource(
            uri="virtual://test/1",
            subdataset=None,
            meta=AuxBandMetadata("object", None, "1", driver_data=prop_req),
            driver_data="test_string",
        )

        srcs = [[aux_source]]
        cfg = AuxLoadParams(
            dtype="object",
            fill_value=None,
            meta=AuxBandMetadata("object", None, "1", driver_data=prop_req),
        )
        used_names = {"time"}
        time_coord = xr.DataArray(
            [np.datetime64("2020-01-01")], dims=["time"], name="time"
        )
        available_coords = {"time": time_coord}
        ctx = None

        result = reader.read(srcs, cfg, used_names, available_coords, ctx)

        assert result.values[0] == "test_string"

    def test_read_with_mixed_data_types(self):
        """Test reading with mixed data types in sources."""
        reader = StacAuxReader()

        prop_req = PropertyLoadRequest(key="test_prop", dtype="float32", units="1")

        aux_source1 = AuxDataSource(
            uri="virtual://test/1",
            subdataset=None,
            meta=AuxBandMetadata("float32", None, "1", driver_data=prop_req),
            driver_data=10.0,
        )
        aux_source2 = AuxDataSource(
            uri="virtual://test/2",
            subdataset=None,
            meta=AuxBandMetadata("float32", None, "1", driver_data=prop_req),
            driver_data="20",  # String instead of float
        )

        srcs = [[aux_source1, aux_source2]]
        cfg = AuxLoadParams(
            dtype="float32",
            fill_value=None,
            meta=AuxBandMetadata("float32", None, "1", driver_data=prop_req),
        )
        used_names = {"time"}
        time_coord = xr.DataArray(
            [np.datetime64("2020-01-01")], dims=["time"], name="time"
        )
        available_coords = {"time": time_coord}
        ctx = None

        result = reader.read(srcs, cfg, used_names, available_coords, ctx)

        # Should handle mixed types gracefully
        assert isinstance(result, xr.DataArray)

    def test_read_with_empty_time_coord(self):
        """Test reading with empty time coordinate."""
        reader = StacAuxReader()

        prop_req = PropertyLoadRequest(key="test_prop", dtype="float32", units="1")

        srcs = []  # No time steps for empty time
        cfg = AuxLoadParams(
            dtype="float32",
            fill_value=-999,
            meta=AuxBandMetadata("float32", -999, "1", driver_data=prop_req),
        )
        used_names = {"time"}
        time_coord = xr.DataArray([], dims=["time"], name="time")
        available_coords = {"time": time_coord}
        ctx = None

        result = reader.read(srcs, cfg, used_names, available_coords, ctx)

        assert result.shape == (0,)

    def test_read_with_missing_time_coord(self):
        """Test reading with missing time coordinate."""
        reader = StacAuxReader()

        prop_req = PropertyLoadRequest(key="test_prop", dtype="float32", units="1")

        aux_source = AuxDataSource(
            uri="virtual://test/1",
            subdataset=None,
            meta=AuxBandMetadata("float32", None, "1", driver_data=prop_req),
            driver_data=42.0,
        )

        srcs = [[aux_source]]
        cfg = AuxLoadParams(
            dtype="float32",
            fill_value=None,
            meta=AuxBandMetadata("float32", None, "1", driver_data=prop_req),
        )
        used_names = {"time"}
        available_coords = {}  # Missing time coordinate
        ctx = None

        with pytest.raises(KeyError):
            reader.read(srcs, cfg, used_names, available_coords, ctx)

    def test_read_with_invalid_cfg(self):
        """Test reading with invalid configuration."""
        reader = StacAuxReader()

        srcs = [[]]
        cfg = AuxLoadParams(
            dtype="float32",
            fill_value=None,
            meta=None,  # Invalid: meta should not be None
        )
        used_names = {"time"}
        time_coord = xr.DataArray(
            [np.datetime64("2020-01-01")], dims=["time"], name="time"
        )
        available_coords = {"time": time_coord}
        ctx = None

        with pytest.raises(AssertionError):
            reader.read(srcs, cfg, used_names, available_coords, ctx)

    def test_read_with_invalid_driver_data(self):
        """Test reading with invalid driver_data in cfg.meta."""
        reader = StacAuxReader()

        srcs = [[]]
        cfg = AuxLoadParams(
            dtype="float32",
            fill_value=None,
            meta=AuxBandMetadata("float32", None, "1", driver_data="invalid"),
        )
        used_names = {"time"}
        time_coord = xr.DataArray(
            [np.datetime64("2020-01-01")], dims=["time"], name="time"
        )
        available_coords = {"time": time_coord}
        ctx = None

        with pytest.raises(AssertionError):
            reader.read(srcs, cfg, used_names, available_coords, ctx)

    def test_read_with_none_fill_value(self):
        """Test reading with None fill value."""
        reader = StacAuxReader()

        prop_req = PropertyLoadRequest(
            key="test_prop", dtype="float32", units="1", nodata=None
        )

        srcs = [[]]  # Empty sources
        cfg = AuxLoadParams(
            dtype="float32",
            fill_value=None,
            meta=AuxBandMetadata("float32", None, "1", driver_data=prop_req),
        )
        used_names = {"time"}
        time_coord = xr.DataArray(
            [np.datetime64("2020-01-01")], dims=["time"], name="time"
        )
        available_coords = {"time": time_coord}
        ctx = None

        # The PropertyLoadRequest.fill_value should handle None nodata
        result = reader.read(srcs, cfg, used_names, available_coords, ctx)
        assert np.isnan(result.values[0])  # Default for float32 when nodata is None


class TestMkParsedItemWithProps:
    """Test cases for mk_parsed_item with props argument."""

    def test_mk_parsed_item_with_simple_values(self):
        """Test mk_parsed_item with simple values."""

        # Test with simple values
        props = {"eo:cloud_cover": 30, "some-other": 42}
        item = mk_parsed_item(bands=[b_("B01")], props=props)

        # Check that auxiliary bands were created
        aux_bands = [(k, v) for k, v in item.bands.items() if k[0] == "_stac_metadata"]
        assert len(aux_bands) == 2

        # Check that driver_data contains the actual values
        for _, source in aux_bands:
            assert isinstance(source, AuxDataSource)
            assert source.meta.driver_data is not None
            assert isinstance(source.meta.driver_data, PropertyLoadRequest)
            assert source.driver_data in [30, 42]  # Actual values

        # Check that aliases were created
        assert "eo_cloud_cover" in item.collection.meta.aliases
        assert "some_other" in item.collection.meta.aliases

    def test_mk_parsed_item_with_tuple_config(self):
        """Test mk_parsed_item with tuple format: (value, config_dict)."""

        # Test with tuple format: (value, config_dict)
        props = {
            "eo:cloud_cover": (
                30,
                {
                    "key": "eo:cloud_cover",
                    "dtype": "float32",
                    "units": "percent",
                    "nodata": -999,
                },
            ),
            "some-other": (
                42,
                {"key": "some-other", "dtype": "int32", "name": "custom_alias"},
            ),
        }
        item = mk_parsed_item(bands=[b_("B01")], props=props)

        # Check that auxiliary bands were created
        aux_bands = [(k, v) for k, v in item.bands.items() if k[0] == "_stac_metadata"]
        assert len(aux_bands) == 2

        # Check that driver_data contains the actual values
        for _, source in aux_bands:
            assert isinstance(source, AuxDataSource)
            assert source.meta.driver_data is not None
            assert isinstance(source.meta.driver_data, PropertyLoadRequest)
            assert source.driver_data in [30, 42]  # Actual values

        # Check that metadata was properly set
        for _, source in aux_bands:
            prop_req = source.meta.driver_data
            if prop_req.key == "eo:cloud_cover":
                assert prop_req.dtype == "float32"
                assert prop_req.units == "percent"
                assert prop_req.nodata == -999
                assert source.driver_data == 30
            elif prop_req.key == "some-other":
                assert prop_req.dtype == "int32"
                assert prop_req.output_name == "custom_alias"
                assert source.driver_data == 42

        # Check aliases
        assert "eo_cloud_cover" in item.collection.meta.aliases
        assert "custom_alias" in item.collection.meta.aliases

    def test_mk_parsed_item_with_mixed_formats(self):
        """Test mk_parsed_item with mixed simple values and tuple configs."""

        # Test with mixed formats
        props = {
            "simple_prop": 25,
            "tuple_prop": (
                0.8,
                {"key": "tuple_prop", "dtype": "float64", "nodata": -999},
            ),
            "another_simple": 100,
        }
        item = mk_parsed_item(bands=[b_("B01")], props=props)

        # Check that auxiliary bands were created
        aux_bands = [(k, v) for k, v in item.bands.items() if k[0] == "_stac_metadata"]
        assert len(aux_bands) == 3

        # Check that driver_data contains the actual values
        for _, source in aux_bands:
            assert isinstance(source, AuxDataSource)
            assert source.meta.driver_data is not None
            assert isinstance(source.meta.driver_data, PropertyLoadRequest)
            assert source.driver_data in [25, 0.8, 100]  # Actual values

        # Check that metadata was properly set
        for _, source in aux_bands:
            prop_req = source.meta.driver_data
            if prop_req.key == "simple_prop":
                assert source.driver_data == 25
            elif prop_req.key == "tuple_prop":
                assert prop_req.dtype == "float64"
                assert prop_req.nodata == -999
                assert source.driver_data == 0.8
            elif prop_req.key == "another_simple":
                assert source.driver_data == 100

        # Check aliases
        assert "simple_prop" in item.collection.meta.aliases
        assert "tuple_prop" in item.collection.meta.aliases
        assert "another_simple" in item.collection.meta.aliases

    def test_mk_parsed_item_without_props(self):
        """Test mk_parsed_item without props (should work as before)."""

        item = mk_parsed_item(bands=[b_("B01")])

        # Should not have any auxiliary bands
        aux_bands = [(k, v) for k, v in item.bands.items() if k[0] == "_stac_metadata"]
        assert len(aux_bands) == 0

        # Should not have any aliases for auxiliary bands
        aux_aliases = [
            k
            for k, v in item.collection.meta.aliases.items()
            if v and v[0][0] == "_stac_metadata"
        ]
        assert len(aux_aliases) == 0

    def test_mk_parsed_item_props_output_names(self):
        """Test that output names are correctly generated from props."""

        props = {
            "cloud.cover": 25,  # Should become "cloud_cover"
            "vegetation:index": 0.8,  # Should become "vegetation_index"
            "temp-data": 15,  # Should become "temp_data"
        }
        item = mk_parsed_item(bands=[b_("B01")], props=props)

        # Check aliases
        assert "cloud_cover" in item.collection.meta.aliases
        assert "vegetation_index" in item.collection.meta.aliases
        assert "temp_data" in item.collection.meta.aliases

    def test_mk_parsed_item_props_metadata(self):
        """Test that auxiliary band metadata is correctly set."""

        props = {"cloud_cover": 25}
        item = mk_parsed_item(bands=[b_("B01")], props=props)

        # Find the auxiliary band
        aux_band = None
        for k, v in item.bands.items():
            if k[0] == "_stac_metadata":
                aux_band = v
                break

        assert aux_band is not None
        assert isinstance(aux_band, AuxDataSource)
        assert aux_band.meta.driver_data is not None
        assert aux_band.meta.driver_data.key == "cloud_cover"
        assert aux_band.driver_data == 25

    def test_mk_parsed_item_raster_group_metadata_includes_aux_bands(self):
        """Test that RasterGroupMetadata includes auxiliary band metadata."""

        props = {
            "cloud_cover": 25,
            "vegetation_index": 0.8,
        }
        item = mk_parsed_item(bands=[b_("B01")], props=props)

        # Check that auxiliary bands are included in the metadata
        aux_bands_in_metadata = {
            k: v
            for k, v in item.collection.meta.bands.items()
            if k[0] == "_stac_metadata"
        }
        assert len(aux_bands_in_metadata) == 2

        # Check that they are AuxBandMetadata objects
        for k, v in aux_bands_in_metadata.items():
            assert isinstance(v, AuxBandMetadata)
            assert v.driver_data is not None
            assert isinstance(v.driver_data, PropertyLoadRequest)

        # Check that raster bands are also included
        raster_bands_in_metadata = {
            k: v
            for k, v in item.collection.meta.bands.items()
            if k[0] != "_stac_metadata"
        }
        assert len(raster_bands_in_metadata) == 1  # B01
        for k, v in raster_bands_in_metadata.items():
            assert isinstance(v, RasterBandMetadata)

        # Check that aux_bands property works
        aux_bands = item.collection.meta.aux_bands
        assert len(aux_bands) == 2
        for k, v in aux_bands.items():
            assert isinstance(v, AuxBandMetadata)

        # Check that raster_bands property works
        raster_bands = item.collection.meta.raster_bands
        assert len(raster_bands) == 1
        for k, v in raster_bands.items():
            assert isinstance(v, RasterBandMetadata)



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.13.0
#   kernelspec:
#     display_name: ODC
#     language: python
#     name: odc
# ---

# %%
import json
from timeit import default_timer as t_now

import geopandas as gpd
import numpy as np
import odc.stac
import planetary_computer as pc
import pystac.item
import pystac_client
from dask.utils import format_bytes
from distributed import Client
from distributed import wait as dask_wait

if "geom_query" in locals():
    bbox = tuple(geom_query.boundingbox)

mode = "site1-tall"
if mode == "site1-tall":
    # mgrs_tiles = ["35MNM", "35LNL", "35LNK", "35LNJ", "35LNH", "35LNG", "35LNF", "35LNE", "35LND"]
    bbox = (27.345815, -14.98724, 27.565542, -7.710992)  # Narrow/Tall epsg:32735
    file_id = "site1-20200606-tall-strip-africa"
    datetime = "2020-06-06"
    query = {}
elif mode == "site2":
    bbox = None
    file_id = "site2-2020_jun_jul-35MNM"
    datetime = "2020-06/2020-07"
    query = {
        "s2:mgrs_tile": {"eq": "35MNM"},
        "s2:nodata_pixel_percentage": {"lt": 10},
    }


cat = pystac_client.Client.open("https://planetarycomputer.microsoft.com/api/stac/v1")
search = cat.search(
    collections=["sentinel-2-l2a"],
    datetime=datetime,
    query=query,
    bbox=bbox,
)
print("Query API end-point")
all_features = search.item_collection_as_dict()

all_features["properties"] = dict(url=search.url, query=search._parameters)
all_features["properties"]

# %%
out_path = Path(f"{file_id}.geojson")
if out_path.exists():
    print(f"File exists, keeping previous version: {out_path}")
else:
    print(f"Writing to: {out_path}")
    json.dump(all_features, open(out_path, "wt"))

# %%
all_items = [pystac.item.Item.from_dict(f) for f in all_features["features"]]

# %%
gdf = gpd.GeoDataFrame.from_features(all_features, "epsg:4326")
display(set(gdf["s2:mgrs_tile"].values), set(gdf.platform), len(set(gdf.datetime)))

_map = gdf.explore(
    "s2:mgrs_tile",
    categorical=True,
    tooltip=[
        "s2:mgrs_tile",
        "datetime",
        "s2:nodata_pixel_percentage",
        "eo:cloud_cover",
    ],
    popup=True,
    style_kwds=dict(fillOpacity=0.0, width=2),
    name="STAC",
)
display(_map)

# %%
display(gdf.head())
# gdf[gdf['s2:nodata_pixel_percentage']>10].explore()

# %%



"""Metadata and data loading model classes."""

from __future__ import annotations

import datetime as dt
import math
from copy import copy
from dataclasses import astuple, dataclass, field, replace
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
)

import numpy as np
from odc.geo import CRS, Geometry, MaybeCRS
from odc.geo.geobox import GeoBox
from odc.geo.types import Unset
from odc.loader.types import (
    AuxBandMetadata,
    AuxDataSource,
    BandIdentifier,
    BandKey,
    BandQuery,
    FixedCoord,
    RasterBandMetadata,
    RasterGroupMetadata,
    RasterSource,
    norm_band_metadata,
    norm_key,
)
from typing_extensions import override


@dataclass(eq=True, frozen=True)
class RasterCollectionMetadata(
    Mapping[BandIdentifier, RasterBandMetadata | AuxBandMetadata]
):
    """
    Information about raster data in a collection.

    We assume that assets with the same names have the same kind of raster data across items within
    a collection. This is built from the combination of data collected from STAC and user
    configuration if supplied.
    """

    name: str
    """Collection name."""

    meta: RasterGroupMetadata
    """
    Band, aliases and extra dimensions metadata.
    """

    has_proj: bool
    """
    Whether to expect/look for ``proj`` extension on item assets.

    Proj data extraction can be disabled by the user with config. It is also disabled if it was not
    detected in the first item.
    """

    band2grid: Dict[str, str]
    """
    Band name to grid name mapping.

    Bands that share the same geometry map to the same grid name. Usually all bands share one common
    grid with the name ``default``. Here again we assume that this grouping of bands to grids is
    stable across the entire collection. This information is used to decide default projection and
    resolution at load time.

    Right now grid information is only extracted from STAC, so any savings from looking up this
    information once across all bands that share common grid is relatively insignificant, but if we
    ever support looking that up from the actual raster data this can speed up the process. This
    also reduces memory pressure somewhat as many bands will share one grid object.
    """

    def band_aliases(self, unique: bool = False) -> Dict[BandKey, List[str]]:
        """
        Compute inverse of alias mapping.

        :return:
          Mapping from canonical name to a list of defined aliases.
        """
        out: Dict[BandKey, List[str]] = {}
        for alias, canon_names in self.meta.aliases.items():
            if unique:
                canon_names = canon_names[:1]

            for cn in canon_names:
                out.setdefault(cn, []).append(alias)
        return out

    def _norm_key(self, k: BandKey) -> str:
        asset, idx = k

        # if single band asset it's just asset name
        if idx == 1 and (asset, 2) not in self.meta.bands and asset != "_stac_metadata":
            return asset

        # if any alias references this key as first choice return that
        for alias, (_k, *_) in self.meta.aliases.items():
            if _k == k:
                return alias

        # Finaly use . notation
        return f"{asset}.{idx}"

    @property
    def all_bands(self) -> List[str]:
        return [self._norm_key(k) for k in self.meta.bands]

    @property
    def prop_bands(self) -> List[str]:
        return [self._norm_key(k) for k in self.meta.bands if k[0] == "_stac_metadata"]

    def normalize_band_query(self, bands: BandQuery = None) -> List[str]:
        if bands is None:
            return self.all_bands

        if isinstance(bands, str):
            bands = [bands]
        elif not isinstance(bands, list):
            bands = list(bands)

        # when subset of raster bands is requested, we still add properties to
        # the query, unless query references at least one property also
        _props = self.prop_bands
        if any(b in _props for b in bands):
            return bands
        return bands + _props

    def resolve_bands(
        self,
        bands: BandQuery = None,
    ) -> Dict[str, RasterBandMetadata | AuxBandMetadata]:
        """
        Query bands taking care of aliases.
        """
        query = self.normalize_band_query(bands)

        return {
            band: self.meta.bands[k]
            for band, k in ((band, self.band_key(band)) for band in query)
        }

    def band_key(self, band: str) -> BandKey:
        """
        Compute canonical band key for an alias/band.

        ``(asset name: str,  band index: int 1..)``
        """
        if (band, 1) in self.meta.bands:
            return (band, 1)

        candidates = self.meta.aliases.get(band, [])
        n = len(candidates)
        if n == 1:
            return candidates[0]
        if n > 1:
            # maybe warn about ambiguity?
            return candidates[0]

        # check if it's asset.<index> form
        parts = band.rsplit(".", 1)
        if len(parts) > 1:
            band, idx = parts
            return (band, int(idx))

        raise ValueError(f"No such band/alias: {band}")

    def canonical_name(self, band: str) -> str:
        """
        Canonical name for an alias.
        """
        return self._norm_key(self.band_key(band))

    @override
    def __getitem__(self, band: BandIdentifier) -> RasterBandMetadata | AuxBandMetadata:
        """
        Query band taking care of aliases.

        :raises: :py:class:`KeyError`
        """
        if isinstance(band, str):
            try:
                band = self.band_key(band)
            except ValueError:
                raise KeyError(band) from None
        return self.meta.bands[band]

    @property
    def bands(self) -> Mapping[BandKey, RasterBandMetadata | AuxBandMetadata]:
        return self.meta.bands

    def meta_for(self, bands: BandQuery = None) -> RasterGroupMetadata:
        """
        Extract raster group metadata for a subset of bands.

        Output uses supplied band names as keys, effectively replacing canonical
        names with aliases supplied by the user.
        """
        return self.meta.patch(
            bands={norm_key(b): self[b] for b in self.normalize_band_query(bands)}
        )

    @property
    def aliases(self) -> Dict[str, List[BandKey]]:
        return self.meta.aliases

    @override
    def __len__(self) -> int:
        return len(self.meta.bands)

    @override
    def __iter__(self) -> Iterator[BandKey]:
        yield from self.meta.bands

    @override
    def __contains__(self, __o: object) -> bool:
        if isinstance(__o, tuple):
            return __o in self.meta.bands
        if isinstance(__o, str):
            return __o in self.meta.aliases or norm_key(__o) in self.meta.bands
        return False

    def __dask_tokenize__(self):
        return astuple(self)

    def patch(self, **kwargs) -> "RasterCollectionMetadata":
        return replace(self, **kwargs)

    def asset_names(self) -> tuple[str, ...]:
        out: list[str] = []
        seen: set[str] = set()
        for asset_name, _ in self.meta.bands:
            if asset_name != "_stac_metadata":
                if asset_name in seen:
                    continue
                seen.add(asset_name)
                out.append(asset_name)

        return tuple(out)


@dataclass(eq=True, frozen=True)
class ParsedItem(Mapping[BandIdentifier, RasterSource | AuxDataSource]):
    """
    Captures essentials parts for data loading from a STAC Item.

    Only includes raster bands of interest.
    """

    # pylint: disable=too-many-instance-attributes

    id: str
    """Item id copied from STAC."""

    collection: RasterCollectionMetadata
    """Collection this Item is part of."""

    bands: Mapping[BandKey, RasterSource | AuxDataSource]
    """Raster bands."""

    geometry: Optional[Geometry] = None
    """Footprint of the dataset."""

    datetime: Optional[dt.datetime] = None
    """Nominal timestamp."""

    datetime_range: Tuple[Optional[dt.datetime], Optional[dt.datetime]] = None, None
    """Time period covered."""

    href: Optional[str] = None
    """Self link from stac item."""

    accessories: dict[str, Any] = field(default_factory=dict)
    """Additional assets"""

    def geoboxes(self, bands: BandQuery = None) -> Tuple[GeoBox, ...]:
        """
        Unique ``GeoBox`` s, highest resolution first.

        :param bands: which bands to consider, default is all
        """
        bands = self.collection.normalize_band_query(bands)

        def _resolution(g: GeoBox) -> float:
            return min(g.resolution.map(abs).xy)

        # TODO: support other geobox types?
        gbx: Set[GeoBox] = set()
        for name in bands:
            b = self.bands.get(self.collection.band_key(name), None)
            if isinstance(b, RasterSource):
                if b.geobox is not None:
                    assert isinstance(b.geobox, GeoBox)
                    gbx.add(b.geobox)

        return tuple(sorted(gbx, key=_resolution))

    def crs(self, bands: BandQuery = None) -> Optional[CRS]:
        """
        First non-null CRS across assets.
        """
        for gbox in self.geoboxes(bands):
            if gbox.crs is not None:
                return gbox.crs

        return None

    def image_geometry(
        self,
        crs: MaybeCRS = Unset(),
        bands: BandQuery = None,
    ) -> Optional[Geometry]:
        """
        Extract footprint of a given band(s) from proj metadata in a given projection.
        """
        if isinstance(crs, Unset):
            crs = None

        for gbox in self.geoboxes(bands):
            if gbox.crs is not None:
                if crs is None or crs == gbox.crs:
                    return gbox.extent
                return gbox.footprint(crs)

        return None

    def safe_geometry(
        self,
        crs: MaybeCRS = Unset(),
        bands: BandQuery = None,
    ) -> Optional[Geometry]:
        """
        Get item geometry footprint in desired projection or native.

        1. Use full-image footprint if proj data is available
        2. Fallback to item geometry if not
        """

        img_geom = self.image_geometry(crs, bands=bands)
        if img_geom is not None:
            return img_geom

        if self.geometry is None:
            return None

        if crs is None or isinstance(crs, Unset):
            return self.geometry

        N = 100  # minimum number of points along perimiter we desire
        min_sample_distance = math.sqrt(self.geometry.area) * 4 / N
        return self.geometry.to_crs(
            crs,
            min_sample_distance,
            check_and_fix=True,
        ).dropna()

    def resolve_bands(
        self, bands: BandQuery = None
    ) -> dict[str, RasterSource | AuxDataSource | None]:
        """
        Query bands taking care of aliases.
        """
        bands = self.collection.normalize_band_query(bands)
        canon = self.collection.band_key

        return {
            k: self.bands.get(_actual, None)
            for k, _actual in ((k, canon(k)) for k in bands)
        }

    @override
    def __getitem__(self, band: BandIdentifier) -> RasterSource | AuxDataSource:
        """
        Query band taking care of aliases.

        :raises: :py:class:`KeyError`
        """
        if isinstance(band, str):
            band = self.collection.band_key(band)
        return self.bands[band]

    @override
    def __len__(self) -> int:
        return len(self.bands)

    @override
    def __iter__(self) -> Iterator[BandKey]:
        yield from self.bands

    @override
    def __contains__(self, k: object) -> bool:
        if isinstance(k, str):
            try:
                return self.collection.band_key(k) in self.bands
            except ValueError:
                return False
        if isinstance(k, tuple):
            return k in self.bands
        return False

    @property
    def nominal_datetime(self) -> dt.datetime:
        """
        Resolve timestamp to a single value.

        - datetime if set
        - start_datetime if set
        - end_datetime if set
        - ``raise ValueError`` otherwise
        """
        for ts in [self.datetime, *self.datetime_range]:
            if ts is not None:
                return ts
        raise ValueError("Timestamp was not populated.")

    @property
    def mid_longitude(self) -> Optional[float]:
        """
        Return longitude of the center point.

        used for "solar day" computation.
        """
        if self.geometry is None:
            return None
        ((lon, _),) = self.geometry.centroid.to_crs("epsg:4326").points
        return lon

    @property
    def solar_date(self) -> dt.datetime:
        """
        Nominal datetime adjusted by longitude.
        """
        lon = self.mid_longitude
        if lon is None:
            return self.nominal_datetime
        return _convert_to_solar_time(self.nominal_datetime, lon)

    def solar_date_at(self, lon: float) -> dt.datetime:
        """
        Nominal datetime adjusted by longitude.
        """
        return _convert_to_solar_time(self.nominal_datetime, lon)

    def strip(self) -> "ParsedItem":
        """
        Copy of self but with stripped bands.
        """
        return replace(
            self,
            bands={k: band.strip() for k, band in self.bands.items()},
            accessories={},
        )

    def assets(self) -> Dict[str, List[RasterSource]]:
        """
        Extract bands grouped by asset they belong to.
        """
        assets: Dict[str, List[Tuple[int, RasterSource]]] = {}
        for (asset, idx), src in self.bands.items():
            if isinstance(src, RasterSource):
                assets.setdefault(asset, []).append((idx, src))

        return {
            k: [src for _, src in sorted(srcs, key=lambda x: x[0])]
            for k, srcs in assets.items()
        }

    @override
    def __hash__(self) -> int:
        return hash((self.id, self.collection.name))

    def __dask_tokenize__(self):
        return (
            self.id,
            self.collection,
            self.bands,
            self.href,
            self.datetime,
            self.datetime_range,
        )


def _default_props_fuser(xx: Sequence[Any]) -> Any:
    n = len(xx)
    if n == 0:
        return None
    if n == 1:
        return xx[0]
    if isinstance(xx[0], str):
        return ",".join((str(x) for x in xx))

    xx = [x for x in xx if isinstance(x, (int, float)) and math.isfinite(x)]
    if len(xx) == 0:
        return None
    if len(xx) == 1:
        return xx[0]
    return sum(xx) / len(xx)


@dataclass(frozen=True)
class PropertyLoadRequest:
    """
    Request to load a property from STAC item as xarray DataArray.

    Attributes:
        key: The key of the property to load from STAC item
        name: Name to use for output DataArray, if None will use the property key
        dtype: Data type to use for loaded data, defaults to float32
    """

    key: str
    name: str | None = None
    dtype: str = "float32"
    nodata: float | None = None
    units: str = "1"
    fuser: Callable[[Sequence[Any]], Any] = _default_props_fuser

    @staticmethod
    def from_user_input(
        inputs: Sequence[str | Mapping[str, Any]],
    ) -> list["PropertyLoadRequest"]:
        """
        Create a list of PropertyLoadRequest objects from user input.

        Args:
            inputs: Sequence of either strings (property keys) or dictionaries with configuration.
                   Dictionaries must have 'key' defined, and can optionally have 'dtype' and 'name'.

        Returns:
            List of PropertyLoadRequest objects

        Raises:
            ValueError: If a dictionary input is missing the required 'key' field
        """

        def _norm(what: str | Mapping[str, Any]) -> "PropertyLoadRequest":
            if isinstance(what, str):
                return PropertyLoadRequest(key=what)
            if isinstance(what, dict):
                if "key" not in what:
                    raise ValueError("Dictionary input must contain 'key' field")
                return PropertyLoadRequest(**what)
            raise ValueError(f"Input must be string or dict, got {type(what)}")

        return [_norm(what) for what in inputs]

    @property
    def output_name(self) -> str:
        if self.name is not None:
            return self.name
        return self.key.replace(".", "_").replace(":", "_").replace("-", "_")

    @property
    def fill_value(self) -> Any:
        dtype = np.dtype(self.dtype)
        if self.nodata is not None:
            return dtype.type(self.nodata)
        if dtype.kind == "f":
            return dtype.type(float("nan"))
        return dtype.type(0)


@dataclass(frozen=True)
class MDParseConfig:
    """Item parsing config."""

    band_defaults: RasterBandMetadata = field(
        default_factory=lambda: norm_band_metadata({})
    )
    band_cfg: Dict[str, RasterBandMetadata] = field(default_factory=dict)
    aliases: Dict[str, BandKey] = field(default_factory=dict)
    ignore_proj: bool = False
    extra_dims: Dict[str, int] = field(default_factory=dict)
    extra_coords: Sequence[FixedCoord] = ()
    with_props: Sequence[PropertyLoadRequest] = field(default_factory=list)

    @staticmethod
    def from_dict(
        cfg: Dict[str, Any], collection_id: str | None = None
    ) -> "MDParseConfig":
        if collection_id is not None:
            if "assets" in cfg:  # Assume it's a single collection config
                _cfg = copy(cfg)
            else:
                _cfg = copy(cfg.get("*", {}))
                _cfg.update(cfg.get(collection_id, {}))
        else:
            _cfg = copy(cfg)

        band_defaults, band_cfg = _norm_band_cfg(_cfg.get("assets", {}))

        aliases = {
            alias: ((band, 1) if isinstance(band, str) else band)
            for alias, band in _cfg.get("aliases", {}).items()
        }
        ignore_proj: bool = _cfg.get("ignore_proj", False)
        extra_dims: Dict[str, int] = _cfg.get("dims", {})
        extra_coords: list[FixedCoord] = []
        cc: dict[str, list[Any]] = _cfg.get("coords", {})
        assert isinstance(cc, dict)
        with_props = _cfg.get("with_properties", cfg.get("with_properties", []))
        assert isinstance(with_props, list)

        for name, val in cc.items():
            assert isinstance(val, list)
            extra_coords.append(FixedCoord(name, val))

        return MDParseConfig(
            band_defaults=band_defaults,
            band_cfg=band_cfg,
            ignore_proj=ignore_proj,
            aliases=aliases,
            extra_dims=extra_dims,
            extra_coords=tuple(extra_coords),
            with_props=PropertyLoadRequest.from_user_input(with_props),
        )


def _norm_band_cfg(
    cfg: Dict[str, Any],
) -> Tuple[RasterBandMetadata, Dict[str, RasterBandMetadata]]:
    fallback = norm_band_metadata(cfg.get("*", {}))
    return fallback, {
        k: norm_band_metadata(v, fallback) for k, v in cfg.items() if k != "*"
    }


def _convert_to_solar_time(utc: dt.datetime, longitude: float) -> dt.datetime:
    # offset_seconds snapped to 1 hour increments
    #    1/15 == 24/360 (hours per degree of longitude)
    offset_seconds = int(longitude / 15) * 3600
    return utc + dt.timedelta(seconds=offset_seconds)



"""STAC Item -> ODC Dataset[eo3]."""

from odc.loader import configure_rio, configure_s3_access
from odc.loader.types import RasterBandMetadata, RasterLoadParams, RasterSource

from ._mdtools import (
    ConversionConfig,
    ParsedItem,
    extract_collection_metadata,
    output_geobox,
    parse_item,
    parse_items,
)
from ._stac_load import load
from .model import RasterCollectionMetadata

stac_load = load


__all__ = (
    "ParsedItem",
    "RasterBandMetadata",
    "RasterCollectionMetadata",
    "RasterLoadParams",
    "RasterSource",
    "ConversionConfig",
    "load",
    "stac_load",
    "configure_rio",
    "configure_s3_access",
    "parse_item",
    "parse_items",
    "extract_collection_metadata",
    "output_geobox",
)


def __dir__():
    return [*__all__, "__version__"]


def __getattr__(name):
    # pylint: disable=import-outside-toplevel
    if name == "__version__":
        from importlib.metadata import version

        return version(__name__)
    raise AttributeError(f"module {__name__} has no attribute {name}")



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.13.0
#   kernelspec:
#     display_name: ODC
#     language: python
#     name: odc
# ---

# %%
import json
import time
from timeit import default_timer as t_now

import datacube
import numpy as np
import odc.stac
import planetary_computer as pc
import pystac.item
import pystac_client
import rioxarray
import stackstac
import xarray as xr
from distributed import Client
from distributed import wait as dask_wait
from odc.stac.bench import (
    BenchLoadParams,
    collect_context_info,
    load_from_json,
    run_bench,
)

# generated by bench-prep-query notebook
fname = [
    "site1-20200606-tall-strip-africa.geojson",
    "site2-2020_jun_jul-35MNM.geojson",
][0]

print(f"Load from file: {fname}")
all_features = json.load(open(fname, "rt"))
print(json.dumps(all_features.get("properties", {}), indent=2))

# %%
params = BenchLoadParams(
    scenario=fname,
    method="stackstac",
    bands=("B02", "B03", "B04"),
    chunks=(2048, 2048),
    patch_url=pc.sign,
    extra={
        "stackstac": {"dtype": "uint16", "fill_value": 0},
        "odc-stac": {"groupby": "solar_day", "stac_cfg": {"*": {"warnings": "ignore"}}},
    },
    # resolution = (2**5)*10,
    # crs="epsg:3857",
)

xx = load_from_json(
    all_features,
    params.with_method("stackstac"),
)

yy = load_from_json(
    all_features,
    params.with_method("odc-stac"),
)

assert xx.spec.transform == xx.geobox.transform
assert xx.geobox == yy.geobox
xx

# %%
client = Client("tcp://127.0.0.1:8786")
if len(client.futures) > 0:
    print("Restarting Client")
    client.restart()
display(client)

# %%
rr_odc, results_odc = run_bench(yy, client, ntimes=3)

# %%
rr_stc, results_stc = run_bench(xx, client, ntimes=3)



"""
Save Landsat 8 pass to GeoTIFF (COG).

This program captures one pass of Band 4 (NIR) of Lansat 8 to a single
cloud optimized GeoTIFF image. Produced image is rotated to maximize
proportion of valid pixels in the result. Data is saved in EPSG:3857 at
native resolution (30m). Produced TIFF is about 4.7GiB.

Data is sourced from Microsoft Planetary Computer:

https://planetarycomputer.microsoft.com/

Python environment

```bash
pip install odc-stac==0.3.0rc1 tqdm planetary_computer pystac-client
```

"""

import planetary_computer
import pystac_client
from affine import Affine
from dask.utils import format_bytes
from odc.geo import geom
from odc.geo.geobox import GeoBox
from tqdm.auto import tqdm

from odc.stac import configure_rio
from odc.stac import load as stac_load

res = 30  # resolution
a = 12.7  # rotation in degrees
band = "SR_B4"

catalog = pystac_client.Client.open(
    "https://planetarycomputer.microsoft.com/api/stac/v1"
)

items = catalog.search(
    collections=["landsat-8-c2-l2"],
    datetime="2021-07-01T08:00:00Z/2021-07-01T09:00:00Z",
    bbox=(-180, -50, 180, 50),
).item_collection()

# Compute Polygon of the pass in EPSG:3857
ls8_pass = geom.unary_union(
    geom.Geometry(item.geometry, "epsg:4326").to_crs("epsg:3857") for item in items
)
assert ls8_pass is not None

# Construct rotated GeoBox
#  rotate geometry
#  construct axis aligned geobox in rotated space
#  then rotate geobox the other way
gbox = Affine.rotation(-a) * GeoBox.from_geopolygon(
    ls8_pass.transform(Affine.rotation(a)),
    resolution=res,
)

# Assume COG datasource, disables looking for external files (it's slow in the cloud)
configure_rio(cloud_defaults=True)

print(f"Loading {band} => {gbox.shape.x:,d}x{gbox.shape.y:,d}")
xx = stac_load(
    items,
    like=gbox,
    bands=[band],
    dtype="int16",
    nodata=0,
    groupby="solar_day",
    resampling="average",
    pool=4,  # Use 4 cores for loading
    progress=tqdm,  #
    patch_url=planetary_computer.sign,
)
print("Load finished")

ts = xx.time[0].dt.strftime("%Y%m%d").item()
fname = f"{band}-{ts}-{res}m.tif"
print(
    f"Will write image to: '{fname}' Raw Size is: {format_bytes(xx[band].data.size*xx[band].dtype.itemsize)}"
)

xx[band].odc.write_cog(
    fname,
    overwrite=True,
    blocksize=2048,
    ovr_blocksize=1024,
    overview_resampling="average",
    intermediate_compression={"compress": "zstd", "zstd_level": 1},
    use_windowed_writes=True,
    compress="zstd",
    zstd_level=6,
    BIGTIFF=True,
    SPARSE_OK=True,
    NUM_THREADS=4,
)



"""stac.load - dc.load from STAC Items."""

from __future__ import annotations

import dataclasses
import functools
import itertools
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    TypeAlias,
    Union,
    cast,
)

import pystac
import pystac.item
import xarray as xr
from dask.utils import ndeepmap
from odc.geo import CRS, MaybeCRS, SomeResolution
from odc.geo.geobox import GeoBox, GeoboxAnchor, GeoboxTiles
from odc.geo.types import Unset
from odc.loader import chunked_load, resolve_chunk_shape, resolve_load_cfg
from odc.loader.types import Band_DType, ReaderDriverSpec

from ._mdtools import ConversionConfig, _resolve_driver, output_geobox, parse_items
from .model import BandQuery, ParsedItem, RasterCollectionMetadata

DEFAULT_CHUNK_FOR_LOAD = 2048
"""Used to partition load when not using Dask."""

GroupbyCallback: TypeAlias = Callable[[pystac.item.Item, ParsedItem, int], Any]

Groupby: TypeAlias = Union[str, GroupbyCallback]


def _collection(items: Iterable[ParsedItem]) -> RasterCollectionMetadata:
    for item in items:
        return item.collection
    raise ValueError("Can't load empty sequence")


def patch_urls(
    item: ParsedItem, edit: Callable[[str], str], bands: BandQuery = None
) -> ParsedItem:
    """
    Map function over dataset measurement urls.

    :param item: Item to edit in place
    :param edit: Function that returns modified url from input url
    :param bands: Only edit specified bands, default is to edit all
    :return: Input item
    """

    if bands is None:
        _bands = {
            k: dataclasses.replace(src, uri=edit(src.uri))
            for k, src in item.bands.items()
        }
    else:
        _to_edit = set(map(item.collection.band_key, bands))
        _bands = {
            k: dataclasses.replace(src, uri=edit(src.uri) if k in _to_edit else src.uri)
            for k, src in item.bands.items()
        }

    return dataclasses.replace(item, bands=_bands)


# pylint: disable=too-many-arguments,too-many-locals,too-many-statements
def load(
    items: Iterable[pystac.item.Item],
    bands: str | Sequence[str] | None = None,
    *,
    groupby: Groupby | None = "time",
    resampling: str | dict[str, str] | None = None,
    dtype: Band_DType = None,
    chunks: dict[str, int | Literal["auto"]] | None = None,
    pool: ThreadPoolExecutor | int | None = None,
    # Geo selection
    crs: MaybeCRS = Unset(),
    resolution: SomeResolution | None = None,
    anchor: GeoboxAnchor | None = None,
    geobox: GeoBox | None = None,
    bbox: tuple[float, float, float, float] | None = None,
    lon: tuple[float, float] | None = None,
    lat: tuple[float, float] | None = None,
    x: tuple[float, float] | None = None,
    y: tuple[float, float] | None = None,
    like: Any = None,
    geopolygon: Any = None,
    intersects: Any = None,
    # UI
    progress: Any = None,
    fail_on_error: bool = True,
    # stac related
    stac_cfg: ConversionConfig | None = None,
    with_properties: Sequence[str | Mapping[str, Any]] | None = None,
    patch_url: Callable[[str], str] | None = None,
    preserve_original_order: bool = False,
    # custom driver
    driver: ReaderDriverSpec | None = None,
    # load behaviour
    fuse_func: str | Mapping[str, str | None] | None = None,
    **kw,
) -> xr.Dataset:
    """
    STAC :class:`~pystac.item.Item` to :class:`xarray.Dataset`.

    Load several STAC :class:`~pystac.item.Item` objects (from the same or similar
    collections) as an :class:`xarray.Dataset`.

    This method can load pixel data directly on a local machine or construct a Dask
    graph that can be processed on a remote cluster.

    .. code-block:: python

       catalog = pystac.Client.open(...)
       query = catalog.search(...)
       xx = odc.stac.load(
           query.items(),
           bands=["red", "green", "blue"],
       )
       xx.red.plot.imshow(col="time")


    :param items:
       Iterable of STAC :class:`~pystac.item.Item` to load

    :param bands:
       List of band names to load, defaults to All. Also accepts
       single band name as input

    .. rubric:: Common Options

    :param groupby:
       Controls what items get placed in to the same pixel plane.

       Following have special meaning:

       * "time" items with exactly the same timestamp are grouped together
       * "solar_day" items captured on the same day adjusted for solar time
       * "id" every item is loaded separately

       Any other string is assumed to be a key in Item's properties dictionary.

       You can also supply custom key function, it should take 3 arguments
       ``(pystac.Item, ParsedItem, index:int) -> Any``

    :param preserve_original_order:
       By default items are sorted by `time, id` within each group to make pixel
       fusing order deterministic. Setting this flag to ``True`` will instead keep
       items within each group in the same order as supplied, so that one can implement
       arbitrary priority for pixel overlap cases.

    :param resampling:
       Controls resampling strategy, can be specified per band

    :param dtype:
       Force output dtype, can be specified per band

    :param chunks:
       Rather than loading pixel data directly, construct
       Dask backed arrays. ``chunks={'x': 2048, 'y': 2048}``

    :param progress:
       Pass in ``tqdm`` progress bar or similar, only used in non-Dask load.

    :param fail_on_error:
        Set this to ``False`` to skip over load failures.

    :param pool:
       Use thread pool to perform load locally, only used in non-Dask load.

    .. rubric:: Control Pixel Grid of Output

    There are many ways to control footprint and resolution of returned data. The most
    precise way is to use :py:class:`~odc.geo.geobox.GeoBox`, ``geobox=GeoBox(..)``.
    Similarly one can use ``like=xx`` to match pixel grid to previously loaded data
    (``xx = odc.stac.load(...)``).

    Other common way is to configure crs and resolution only

    .. code-block:: python

       xx = odc.stac.load(...
           crs="EPSG:3857",
           resolution=10)

       # resolution units must match CRS
       # here we assume 1 degree == 111km to load at roughly
       # the same 10m resolution as statement above.
       yy = odc.stac.load(...
           crs="EPSG:4326",
           resolution=0.00009009)

    By default :py:func:`odc.stac.load` loads all available pixels in the requested
    projection and resolution. To limit extent of loaded data you have to supply bounds via
    either ``geobox=`` or ``like=`` parameters (these also select projection and resolution).
    Alternatively use a pair of ``x, y`` or ``lon, lat`` parameters. ``x, y`` allows you to
    specify bounds in the output projection, while ``lon, lat`` operate in degrees. You can also
    use ``bbox`` which is equivalent to ``lon, lat``.

    It should be noted that returned data is likely to reach outside of the requested bounds by
    fraction of a pixel when using ``bbox``, ``x, y`` or ``lon, lat`` mechanisms. This is due to
    pixel grid "snapping". Pixel edges will still start at ``N*pixel_size`` where ``N is int``
    regardless of the requested bounding box.

    :param crs:
       Load data in a given CRS. Special name of ``"utm"`` is also understood, in which case an
       appropriate UTM projection will be picked based on the output bounding box.

    :param resolution:
       Set resolution of output in units of the output CRS. This can be a single float, in which
       case pixels are assumed to be square with ``Y`` axis flipped. To specify non-square or
       non-flipped pixels use :py:func:`odc.geo.resxy_` or :py:func:`odc.geo.resyx_`.
       ``resolution=10`` is equivalent to ``resolution=odc.geo.resxy_(10, -10)``.

       Resolution must be supplied in the units of the output CRS. Units are commonly meters
       for *Projected* and degrees for *Geographic* CRSs.

    :param bbox:
       Specify bounding box in Lon/Lat. ``[min(lon), min(lat), max(lon), max(lat)]``

    :param lon:
       Define output bounds in Lon/Lat
    :param lat:
       Define output bounds in Lon/Lat

    :param x:
       Define output bounds in output projection coordinate units
    :param y:
       Define output bounds in output projection coordinate units

    :param anchor:
       Controls pixel snapping, default is to align pixel grid to ``X``/``Y``
       axis such that pixel edges align with ``x=0, y=0``. Other common option is to
       align pixel centers to ``0,0`` rather than edges.

    :param geobox:
       Allows to specify exact region/resolution/projection using
       :class:`~odc.geo.geobox.GeoBox` object

    :param like:
       Match output grid to the data loaded previously.

    :param geopolygon:
       Limit returned result to a bounding box of a given geometry. This could be an
       instance of :class:`~odc.geo.geom.Geometry`, GeoJSON dictionary,
       GeoPandas DataFrame, or any object implementing ``__geo_interface__``. We assume
       ``EPSG:4326`` projection for dictionary and Shapely inputs. CRS information available
       on GeoPandas inputs should be understood correctly.

    :param intersects:
       Simple alias to `geopolygon` so that the same inputs work for `pystac_client.Client.search`
       as they do here.

    .. rubric:: STAC Related Options

    :param stac_cfg: Controls interpretation of :py:class:`pystac.Item`. Mostly used to specify "missing"
       metadata like pixel data types.

    :param with_properties:
       List of properties to load from STAC item. Can be a list of strings or dictionaries with
       the following fields: ``.key``, ``.name``, ``.dtype``, ``.nodata``, ``.units``, ``.fuser``.

    :param patch_url:
       Optionally transform url of every band before loading

    .. rubric:: Load behaviour options

    :param driver:
       Optional. If provided, use the specified driver to load the data.

    :param fuse_func:
        Function used to fuse/combine/reduce data with the ``group_by`` parameter.

        By default, pixels are only copied where valid (i.e. not nodata) pixels
        have not yet been copied from previous items.

        If data (especially categorical data) appears wrong or unexpected in areas
        where items overlap, then an appropriate fuse_func may help.

        The fuse_func can perform specific combining steps and can be specified per band.

    .. rubric:: Custom fuser functions

    Custom fuse functions should be defined as follows:

    .. code-block:: python

            def my_fuser(dst: np.ndarray, src: np.ndarray) -> None:
                # Create a boolean mask array of pixels from this src array to copy.
                mask = pixels_to_copy(src)

                # Efficiently copy only masked pixels to dst.
                np.copyto(dst, src, where=mask)

    For an example of a more sophisticated fuser function, see
    https://github.com/GeoscienceAustralia/dea-notebooks/blob/77e9e3a05c104f4a0de91857905acce5853975b6/Tools/dea_tools/datahandling.py#L713

    Fuser functions are passed to odc-stac as importable strings (fully qualified Python
    names of top-level functions) so that they can be serialised to dask workers.

    In the following example, the ``my_fuser`` function is used for ``band0``, the default nodata-only
    fuser is used for ``band1`` and the ``other_fuser`` function is used for all other raster bands:

    .. code-block:: python

        data = odc.stac.load(...,
            fuse_func={
                "band0": "mymodule.my_fuser",
                "band1": None,
                "*": "mymodule.other_fuser",
            }
        )

    :return:
       :py:class:`xarray.Dataset` with requested bands populated


    .. rubric:: Complete Example Code

    .. code-block:: python

       import planetary_computer as pc
       from pystac_client import Client

       from odc import stac

       catalog = Client.open("https://planetarycomputer.microsoft.com/api/stac/v1")
       query = catalog.search(
           collections=["sentinel-2-l2a"],
           datetime="2019-06-06",
           query={"s2:mgrs_tile": dict(eq="06VVN")},
       )

       xx = stac.load(
           query.items(),
           bands=["red", "green", "blue"],
           resolution=100,  # 1/10 of the native 10m resolution
           patch_url=pc.sign,
       )
       xx.red.plot.imshow(col="time", size=8, aspect=1)


    .. rubric:: Example Optional Configuration

    Sometimes data source might be missing some optional STAC extensions. With ``stac_cfg=`` parameter
    one can supply that information at load time. Configuration is per collection per asset. You can
    provide information like pixel data type, ``nodata`` value used, ``unit`` attribute and band aliases
    you would like to use.

    Sample ``stac_cfg={..}`` parameter:

    .. code-block:: yaml

       sentinel-2-l2a:  # < name of the collection, i.e. ``.collection_id``
         assets:
           "*":  # Band named "*" contains band info for "most" bands
             data_type: uint16
             nodata: 0
             unit: "1"
           SCL:  # Those bands that are different than "most"
             data_type: uint8
             nodata: 0
             unit: "1"
         aliases:  #< unique alias -> canonical map
           rededge: B05
           rededge1: B05
           rededge2: B06
           rededge3: B07

       some-other-collection:
         assets:
         #...

       "*": # Applies to all collections if not defined on a collection
         warnings: ignore  # ignore|all (default all)

    """
    # pylint: disable=unused-argument,too-many-branches
    if bands is None:
        # dc.load name for bands is measurements
        bands = kw.pop("measurements", None)

    # normalize args
    # dc.load compatible name for crs is `output_crs`
    if isinstance(crs, Unset) or crs is None:
        crs = cast(MaybeCRS, kw.pop("output_crs", None))

    if groupby is None:
        groupby = "id"

    rdr, md_parser = _resolve_driver(driver, stac_cfg, with_properties=with_properties)

    items = list(items)
    _parsed = list(parse_items(items, md_plugin=md_parser))

    # Check we have all the bands of interest
    # will raise ValueError if no such band/alias
    collection = _collection(_parsed)
    bands_to_load = collection.resolve_bands(bands)
    bands = list(bands_to_load)

    load_cfg = resolve_load_cfg(
        bands_to_load,
        resampling,
        dtype=dtype,
        use_overviews=kw.get("use_overviews", True),
        nodata=kw.get("nodata", None),
        fail_on_error=fail_on_error,
        fuse_func=fuse_func,
    )
    if patch_url is not None:
        _parsed = [patch_urls(item, edit=patch_url, bands=bands) for item in _parsed]

    if geopolygon is None and intersects is not None:
        geopolygon = intersects

    gbox = output_geobox(
        _parsed,
        bands=bands,
        crs=crs,
        resolution=resolution,
        anchor=anchor,
        align=kw.get("align", None),
        geobox=geobox,
        like=like,
        geopolygon=geopolygon,
        bbox=bbox,
        lon=lon,
        lat=lat,
        x=x,
        y=y,
    )

    if gbox is None:
        # TODO: handle no raster bands case here by creating some fake
        # geobox when only aux bands are present/requested for loading
        raise ValueError("Failed to auto-guess CRS/resolution.")

    # Time dimension
    ((mid_lon, _),) = gbox.extent.centroid.to_crs("epsg:4326").points
    _grouped_idx = _group_items(
        items,
        _parsed,
        groupby,
        mid_lon,
        preserve_original_order=preserve_original_order,
    )

    tss = _extract_timestamps(ndeepmap(2, lambda idx: _parsed[idx], _grouped_idx))
    meta = collection.meta_for(bands)

    if chunks is not None:
        chunk_shape = resolve_chunk_shape(
            len(tss),
            gbox,
            chunks,
            cfg=load_cfg,
            extra_dims=meta.extra_dims_full(),
        )
    else:
        chunk_shape = (1, DEFAULT_CHUNK_FOR_LOAD, DEFAULT_CHUNK_FOR_LOAD)

    # Spatio-temporal binning
    assert isinstance(gbox.crs, CRS)
    gbt = GeoboxTiles(gbox, (chunk_shape[1], chunk_shape[2]))
    tyx_bins = dict(_tyx_bins(_grouped_idx, _parsed, gbt))
    srcs = [item.resolve_bands(bands) for item in _parsed]
    debug = kw.get("debug", False)

    def _with_debug_info(ds: xr.Dataset, **kw) -> xr.Dataset:
        # expose data for debugging
        if not debug:
            return ds

        from types import SimpleNamespace  # pylint: disable=import-outside-toplevel

        ds.encoding.update(
            debug=SimpleNamespace(
                gbt=gbt,
                mid_lon=mid_lon,
                parsed=_parsed,
                srcs=srcs,
                grouped_idx=_grouped_idx,
                tyx_bins=tyx_bins,
                bands_to_load=bands_to_load,
                load_cfg=load_cfg,
                **kw,
            )
        )
        return ds

    rdr_env = rdr.capture_env()
    return _with_debug_info(
        chunked_load(
            load_cfg,
            meta,
            srcs,
            tyx_bins,
            gbt,
            tss,
            rdr_env,
            rdr,
            chunks=chunks,
            pool=pool,
            progress=progress,
            dtype=dtype,
        )
    )


def _extract_timestamps(grouped: List[List[ParsedItem]]) -> List[datetime]:
    def _ts(group: List[ParsedItem]) -> datetime:
        assert len(group) > 0
        return group[0].nominal_datetime.replace(tzinfo=None)

    return list(map(_ts, grouped))


# pylint: disable=unused-argument
def _groupby_solar_day(
    item: pystac.item.Item,
    parsed: ParsedItem,
    idx: int,
    lon: Optional[float] = None,
):
    if lon is None:
        return parsed.solar_date.date()
    return parsed.solar_date_at(lon).date()


def _groupby_time(
    item: pystac.item.Item,
    parsed: ParsedItem,
    idx: int,
):
    return parsed.nominal_datetime


def _groupby_id(
    item: pystac.item.Item,
    parsed: ParsedItem,
    idx: int,
):
    return idx


def _groupby_property(
    item: pystac.item.Item,
    parsed: ParsedItem,
    idx: int,
    key: str = "",
):
    return item.properties.get(key, None)


def _resolve_groupby(groupby: Groupby, lon: Optional[float] = None) -> GroupbyCallback:
    if not isinstance(groupby, str):
        return groupby
    if groupby == "time":
        return _groupby_time
    if groupby == "solar_day":
        return functools.partial(_groupby_solar_day, lon=lon)
    if groupby == "id":
        return _groupby_id

    return functools.partial(_groupby_property, key=groupby)


def _group_items(
    items: List[pystac.item.Item],
    parsed: List[ParsedItem],
    groupby: Groupby,
    lon: Optional[float] = None,
    preserve_original_order=False,
) -> List[List[int]]:
    assert len(items) == len(parsed)

    group_key = _resolve_groupby(groupby, lon=lon)

    def _sorter(idx: int):
        _group = group_key(items[idx], parsed[idx], idx)

        if preserve_original_order:
            # Sort by group_key but keeping original item order within each group
            return (_group, idx)

        # Sort by group_key, but then time,id within each group
        return (_group, parsed[idx].nominal_datetime, parsed[idx].id)

    ii = sorted(range(len(parsed)), key=_sorter)

    return [
        list(group)
        for _, group in itertools.groupby(
            ii, lambda idx: group_key(items[idx], parsed[idx], idx)
        )
    ]


def _tiles(item: ParsedItem, gbt: GeoboxTiles) -> Iterator[Tuple[int, int]]:
    geom = item.safe_geometry(gbt.base.crs)
    if geom is None:
        raise ValueError("Can not process items without defined footprint")
    yield from gbt.tiles(geom)


def _tyx_bins(
    grouped: List[List[int]],
    items: List[ParsedItem],
    gbt: GeoboxTiles,
) -> Iterator[Tuple[Tuple[int, int, int], List[int]]]:
    for t_idx, group in enumerate(grouped):
        _yx: Dict[Tuple[int, int], List[int]] = {}

        for item_idx in group:
            for idx in _tiles(items[item_idx], gbt):
                _yx.setdefault(idx, []).append(item_idx)

        yield from (((t_idx, *idx), ii_item) for idx, ii_item in _yx.items())



"""CLI app for benchmarking."""

import json
from datetime import datetime
from time import sleep
from typing import Any, Dict, Optional

import click
import distributed
import rasterio.enums

from odc.stac.bench import (
    SAMPLE_SITES,
    BenchLoadParams,
    dump_site,
    load_from_json,
    load_results,
    run_bench,
)

# pylint: disable=too-many-arguments,too-many-locals

RIO_RESAMPLING_NAMES = [it.name for it in rasterio.enums.Resampling]


@click.group("odc-stac-bench")
def main() -> None:
    """Benchmarking tool for odc.stac."""


@main.command("prepare")
@click.option("--sample-site", type=str, help="Use one of sample sites")
@click.option(
    "--list-sample-sites",
    is_flag=True,
    default=False,
    help="Print available sample sites",
)
@click.option(
    "--from-file",
    help="From json config file",
    type=click.Path(exists=True, dir_okay=False, readable=True),
)
@click.option("--overwrite", is_flag=True, help="Overwite output file")
def prepare(sample_site: str, list_sample_sites: bool, from_file, overwrite):
    """Prepare benchmarking dataset."""
    if list_sample_sites:
        click.echo("Sample sites:")
        for site_name in SAMPLE_SITES:
            click.echo(f"   {site_name}")
        return

    site: Optional[Dict[str, Any]] = None
    if sample_site is not None:
        site = SAMPLE_SITES.get(sample_site, None)
        if site is None:
            raise click.ClickException(f"No such site: {sample_site}")
        print("Site config:")
        print("------------------------------------------")
        print(json.dumps(site, indent=2))
        print("------------------------------------------")
    elif from_file is not None:
        with open(from_file, "rt", encoding="utf8") as src:
            site = json.load(src)

    if site is None:
        raise click.ClickException("Have to supply one of --sample-site or --from-file")
    dump_site(site, overwrite=overwrite)


@main.command("dask")
@click.option(
    "--n-workers", type=int, default=1, help="Number of workers to launch (1)"
)
@click.option(
    "--threads-per-worker", type=int, help="Number of threads per worker (all cpus)"
)
@click.option("--memory-limit", type=str, help="Configure worker memory limit")
def _dask(n_workers, threads_per_worker, memory_limit) -> None:
    """Launch local Dask Cluster."""
    client = distributed.Client(
        n_workers=n_workers,
        threads_per_worker=threads_per_worker,
        memory_limit=memory_limit,
    )
    info = client.scheduler_info()
    print(f"Launched Dask Cluster: {info['address']}")
    print(f"   --scheduler='{info['address']}'")
    while True:
        try:
            sleep(1)
        except KeyboardInterrupt:
            print("Terminating")
            client.shutdown()
            return


@main.command("run")
@click.option(
    "--config",
    "-c",
    type=click.Path(exists=True, dir_okay=False, readable=True),
    required=False,
    help="Experiment configuration in json format",
)
@click.option(
    "--ntimes", "-n", type=int, default=1, help="Configure number of times to run"
)
@click.option(
    "--method",
    help="Data loading method",
    type=click.Choice(["odc-stac", "stackstac"]),
)
@click.option("--bands", type=str, help="Comma separated list of bands")
@click.option("--chunks", type=int, help="Chunk size Y,X order", nargs=2)
@click.option("--resolution", type=float, help="Set output resolution")
@click.option("--crs", type=str, help="Set CRS")
@click.option(
    "--resampling",
    help="Resampling method when changing resolution/projection",
    type=click.Choice(RIO_RESAMPLING_NAMES),
)
@click.option("--show-config", is_flag=True, help="Show configuration only, don't run")
@click.option(
    "--scheduler", default="tcp://localhost:8786", help="Dask server to connect to"
)
@click.argument("site", type=click.Path(exists=True, dir_okay=False, readable=True))
def run(
    site: str,
    config,
    method: str,
    ntimes: int,
    bands: str,
    chunks,
    resolution: float,
    crs: str,
    resampling: str,
    show_config: bool,
    scheduler,
) -> None:
    """
    Run data load benchmark using Dask.

    SITE is a GeoJSON file produced by `prepare` step.
    """
    cfg: Optional[BenchLoadParams] = None
    if config is not None:
        with open(config, "rt", encoding="utf8") as src:
            cfg = BenchLoadParams.from_json(src.read())
    else:
        cfg = BenchLoadParams(
            method="odc-stac",
            chunks=(2048, 2048),
            extra={
                "stackstac": {"dtype": "uint16", "fill_value": 0},
                "odc-stac": {
                    "groupby": "solar_day",
                    "stac_cfg": {"*": {"warnings": "ignore"}},
                },
            },
        )

    if chunks:
        cfg.chunks = chunks
    if method is not None:
        cfg.method = method
    if bands is not None:
        cfg.bands = tuple(bands.split(","))
    if resolution is not None:
        cfg.resolution = resolution
    if crs is not None:
        cfg.crs = crs
    if resampling is not None:
        cfg.resampling = resampling
    if not cfg.scenario:
        cfg.scenario = site.rsplit(".", 1)[0]

    with open(site, "rt", encoding="utf8") as src:
        site_geojson = json.load(src)

    print(f"Loaded: {len(site_geojson['features'])} STAC items from '{site}'")

    print("Will use following load configuration")
    print("-" * 60)
    print(cfg.to_json(indent=2))
    print("-" * 60)

    if show_config:
        return

    print(f"Connecting to Dask Scheduler: {scheduler}")
    client = distributed.Client(scheduler)

    print("Constructing Dask graph")
    xx = load_from_json(site_geojson, cfg)
    print(f"Starting benchmark run ({ntimes} runs)")
    print("=" * 60)

    ts = datetime.now().strftime("%Y%m%dT%H%M%S.%f")
    results_file = f"{cfg.scenario}_{ts}.pkl"
    print(f"Will write results to: {results_file}")
    _ = run_bench(xx, client, ntimes=ntimes, results_file=results_file)
    print("=" * 60)
    print("Finished")


@main.command("report")
@click.option(
    "--matching", type=str, help="Supply glob pattern instead of individual .pkl files"
)
@click.option(
    "--output",
    type=str,
    help="File to write CSV data, if not supplied will write to stdout",
)
@click.argument(
    "pkls", type=click.Path(exists=True, dir_okay=False, readable=True), nargs=-1
)
def report(matching: str, output: str, pkls) -> None:
    """
    Collate results of multiple benchmark experiments.

    Read pickle files produced by the `run` command and assemble
    them into one CSV file.
    """
    if matching is not None:
        data_raw = load_results(matching)
    else:
        data_raw = load_results(pkls)

    if output is None:
        print(data_raw.to_csv())
    else:
        data_raw.to_csv(output)



"""
Common test data and utilities for STAC tests.
"""

from pystac import Item

from odc.stac._mdtools import RasterBandMetadata

# fmt: off
S2_ALL_BANDS = {
    "B01", "B02", "B03", "B04", "B05", "B06",
    "B07", "B08", "B09", "B11", "B12", "B8A",
    "AOT", "SCL", "WVP", "visual",
}
# fmt: on


STAC_CFG = {
    "sentinel-2-l2a": {
        "assets": {
            "*": RasterBandMetadata("uint16", 0, "1"),
            "SCL": RasterBandMetadata("uint8", 0, "1"),
            "visual": {"data_type": "uint8", "nodata": 0, "unit": "1"},
        },
        "aliases": {
            # Work around duplicate rededge common_name
            # by defining custom unique aliases
            "rededge1": "B05",
            "rededge2": "B06",
            "rededge3": "B07",
        },
    }
}

NO_WARN_CFG = {"*": {"warnings": "ignore"}}


def mk_stac_item(
    _id, datetime="2012-12-12T00:00:00Z", geometry=None, stac_extensions=None, **props
):
    if stac_extensions is None:
        stac_extensions = []

    return Item.from_dict(
        {
            "type": "Feature",
            "stac_version": "1.0.0",
            "id": str(_id),
            "properties": {
                "datetime": datetime,
                **props,
            },
            "geometry": geometry,
            "links": [],
            "assets": {},
            "stac_extensions": stac_extensions,
        }
    )



# pylint: disable=redefined-outer-name,missing-module-docstring,missing-function-docstring,missing-class-docstring
# pylint: disable=use-implicit-booleaness-not-comparison,protected-access
from __future__ import annotations

import pystac
import pystac.asset
import pystac.collection
import pystac.item
import pystac.utils
import pytest
from odc.geo import geom
from odc.geo.geobox import AnchorEnum, GeoBox, geobox_union_conservative
from odc.geo.xr import xr_zeros
from odc.loader.testing.fixtures import FakeMDPlugin
from odc.loader.types import (
    FixedCoord,
    RasterBandMetadata,
    RasterGroupMetadata,
    RasterSource,
)
from pystac.extensions.projection import ProjectionExtension

from odc.stac._mdtools import (
    MDParseConfig,
    _auto_load_params,
    _most_common_gbox,
    _normalize_geometry,
    asset_geobox,
    band_metadata,
    compute_eo3_grids,
    extract_collection_metadata,
    has_proj_ext,
    has_raster_ext,
    is_raster_data,
    output_geobox,
    parse_item,
    parse_items,
)
from odc.stac.model import ParsedItem
from odc.stac.testing.stac import b_, mk_parsed_item, to_stac_item

from .common import NO_WARN_CFG, S2_ALL_BANDS, STAC_CFG

GBOX = GeoBox.from_bbox((-20, -10, 20, 10), "epsg:3857", shape=(200, 400))


def test_mdparse_config() -> None:
    assert MDParseConfig() == MDParseConfig()
    assert MDParseConfig.from_dict({}) == MDParseConfig()
    assert MDParseConfig.from_dict({}, "cc") == MDParseConfig()
    assert MDParseConfig().extra_coords == ()
    assert MDParseConfig().extra_dims == {}

    cfg = {
        "assets": {"visual": {"data_type": "uint8", "dims": ["y", "x", "rgb"]}},
        "dims": {"rgb": 3},
        "coords": {"rgb": ["r", "g", "b"]},
    }
    assert MDParseConfig.from_dict(cfg) == MDParseConfig.from_dict({"*": cfg}, "cc")
    assert MDParseConfig.from_dict(cfg) == MDParseConfig.from_dict({"cc": cfg}, "cc")
    assert MDParseConfig() == MDParseConfig.from_dict({"cc": cfg}, "not-in-cfg")
    assert MDParseConfig.from_dict(cfg) == MDParseConfig.from_dict(cfg, "irrelevant")

    cfg = MDParseConfig.from_dict(cfg)
    assert cfg.extra_dims == {"rgb": 3}
    assert cfg.extra_coords == (FixedCoord("rgb", ["r", "g", "b"]),)


def test_is_raster_data(sentinel_stac_ms: pystac.item.Item) -> None:
    item = sentinel_stac_ms
    assert "B01" in item.assets
    assert "B02" in item.assets

    assert is_raster_data(item.assets["B01"])

    # check case when roles are missing
    item.assets["B02"].roles = None
    assert is_raster_data(item.assets["B02"])


def test_eo3_grids(sentinel_stac_ms: pystac.item.Item) -> None:
    item0 = sentinel_stac_ms

    item = item0.clone()
    assert item.collection_id == "sentinel-2-l2a"

    data_bands = {
        name: asset
        for name, asset in item.assets.items()
        if is_raster_data(asset, check_proj=True)
    }

    grids, b2g = compute_eo3_grids(data_bands)
    assert set(grids) == set("default g20 g60".split(" "))
    assert set(grids) == set(b2g.values())
    assert set(b2g) == set(data_bands)

    # test the case where there are different shapes for the same gsd
    # clashing grid names should be resolved
    ProjectionExtension.ext(item.assets["B01"]).shape = (100, 200)  # type: ignore
    grids, b2g = compute_eo3_grids(data_bands)
    assert b2g["B01"] != b2g["B02"]

    # More than 1 CRS should work
    item = item0.clone()
    ProjectionExtension.ext(item.assets["B01"]).epsg = 3857
    assert ProjectionExtension.ext(item.assets["B01"]).epsg == 3857

    data_bands = {k: item.assets[k] for k in data_bands}
    grids, b2g = compute_eo3_grids(data_bands)
    assert b2g["B01"] != b2g["B02"]
    crs = grids[b2g["B01"]].crs
    assert crs is not None
    assert crs.epsg == 3857


def test_asset_geobox(sentinel_stac: pystac.item.Item) -> None:
    item0 = sentinel_stac
    item = item0.clone()
    asset = item.assets["B01"]
    geobox = asset_geobox(asset)
    assert geobox.shape == (1830, 1830)

    # Tests non-affine transofrm ValueError
    item = item0.clone()
    asset = item.assets["B01"]
    ProjectionExtension.ext(asset).transform[-1] = 2  # type: ignore
    with pytest.raises(ValueError):
        asset_geobox(asset)

    # Tests wrong-sized transform transofrm ValueError
    item = item0.clone()
    asset = item.assets["B01"]
    ProjectionExtension.ext(asset).transform = [1, 1, 2]
    with pytest.raises(ValueError):
        asset_geobox(asset)

    # Test missing transform transofrm ValueError
    item = item0.clone()
    asset = item.assets["B01"]
    ProjectionExtension.ext(asset).transform = None
    with pytest.raises(ValueError):
        asset_geobox(asset)

    # Test no proj extension case
    item = item0.clone()
    item.stac_extensions = []
    asset = item.assets["B01"]
    with pytest.raises(ValueError):
        asset_geobox(asset)


def test_has_proj_ext(sentinel_stac_ms_no_ext: pystac.item.Item) -> None:
    assert has_proj_ext(sentinel_stac_ms_no_ext) is False


def test_band_metadata(sentinel_stac_ms_with_raster_ext: pystac.item.Item) -> None:
    item = sentinel_stac_ms_with_raster_ext.clone()
    assert has_raster_ext(item) is True
    asset = item.assets["SCL"]
    bm = band_metadata(asset, RasterBandMetadata("uint16", 0, "1"))
    assert bm == [RasterBandMetadata("uint8", 0, "1")]

    # Test multiple bands per asset produce multiple outputs
    asset.extra_fields["raster:bands"].append({"nodata": -10})
    bm = band_metadata(asset, RasterBandMetadata("uint16", 0, "1"))
    assert bm == [
        RasterBandMetadata("uint8", 0, "1"),
        RasterBandMetadata(data_type="uint16", nodata=-10, units="1"),
    ]


def test_is_raster_data_more() -> None:
    def _a(href="http://example.com/", **kw):
        return pystac.asset.Asset(href, **kw)

    assert is_raster_data(_a(media_type="image/jpeg")) is True
    assert is_raster_data(_a(media_type="image/jpeg", roles=["data"])) is True
    assert is_raster_data(_a(media_type="image/jpeg", roles=["overview"])) is False
    assert is_raster_data(_a(media_type="image/jpeg", roles=["thumbnail"])) is False

    # no media type defined
    assert is_raster_data(_a(roles=["data"])) is True
    assert is_raster_data(_a(roles=["metadata"])) is False
    assert is_raster_data(_a(roles=["custom-22"])) is False

    # based on extension
    assert is_raster_data(_a(href="/foo.tif")) is True
    assert is_raster_data(_a(href="/foo.tiff")) is True
    assert is_raster_data(_a(href="/foo.TIF")) is True
    assert is_raster_data(_a(href="/foo.TIFF")) is True
    assert is_raster_data(_a(href="/foo.jpeg")) is True
    assert is_raster_data(_a(href="/foo.jpg")) is True


def test_extract_md(sentinel_stac_ms: pystac.item.Item) -> None:
    item0 = sentinel_stac_ms
    item = pystac.Item.from_dict(item0.to_dict())

    assert item.collection_id in STAC_CFG

    md = extract_collection_metadata(item, STAC_CFG)

    assert md.name == "sentinel-2-l2a"

    assert set(md.all_bands) == S2_ALL_BANDS

    # check defaults were set
    for b in ["B01", "B02", "B03"]:
        bk: tuple[str, int] = (b, 1)
        assert md.bands[bk].data_type == "uint16"
        assert md.bands[bk].nodata == 0
        assert md.bands[bk].units == "1"

    assert md.bands[("SCL", 1)].data_type == "uint8"
    assert md.bands[("visual", 1)].data_type == "uint8"

    # check aliases configuration
    assert md.aliases["rededge"] == [("B05", 1), ("B06", 1), ("B07", 1), ("B8A", 1)]
    assert md.aliases["rededge1"] == [("B05", 1)]
    assert md.aliases["rededge2"] == [("B06", 1)]
    assert md.aliases["rededge3"] == [("B07", 1)]

    # check without config
    md = extract_collection_metadata(item)

    for band in md.bands.values():
        assert band.data_type == "float32"
        assert band.nodata is None
        assert band.units == "1"

    # Test that multiple CRSs per item work
    item = pystac.Item.from_dict(item0.to_dict())
    ProjectionExtension.ext(item.assets["B01"]).epsg = 3857
    assert ProjectionExtension.ext(item.assets["B01"]).crs_string == "EPSG:3857"
    md = extract_collection_metadata(item, NO_WARN_CFG)
    assert md.band2grid["B01"] != md.band2grid["B02"]

    # Test no-collection name item
    item = pystac.Item.from_dict(item0.to_dict())
    item.collection_id = None
    md = extract_collection_metadata(item, NO_WARN_CFG)
    assert md.name == "_"


def test_parse_item_with_plugin() -> None:
    item = pystac.item.Item.from_dict(
        {
            "type": "Feature",
            "id": "some-item",
            "collection": "c",
            "properties": {"datetime": "2022-02-02T00:00:00Z"},
            "geometry": None,
            "links": [],
            "assets": {
                "AA": {
                    "href": "http://example.com/items/b1.nc",
                    "type": "application/x-netcdf",
                    "roles": ["data"],
                }
            },
            "stac_version": "1.0.0",
            "stac_extensions": [""],
        }
    )
    group_md = RasterGroupMetadata(
        bands={
            ("AA", 1): RasterBandMetadata("uint8", 0, "1", dims=("y", "x", "b")),
            ("AA", 2): RasterBandMetadata("float32"),
        },
        aliases={"b1": [("AA", 1)], "b2": [("AA", 2)]},
        extra_dims={"b": 3},
        extra_coords=[FixedCoord("b", ["r", "g", "b"])],
    )
    md_plugin = FakeMDPlugin(group_md, {"foo": "bar"})

    pit = parse_item(item, md_plugin=md_plugin)
    assert isinstance(pit, ParsedItem)
    assert pit.collection["b1"].data_type == "uint8"
    assert pit.collection["b1"].dims == ("y", "x", "b")
    assert pit.collection["b2"].data_type == "float32"
    assert pit["b1"].driver_data == {"foo": "bar"}
    assert pit["b2"].driver_data == {"foo": "bar"}
    assert pit["b1"].subdataset is None
    assert pit["b2"].subdataset is None

    md_plugin = FakeMDPlugin(group_md, {"foo": "bar"}, add_subdataset=True)
    pit = parse_item(item, md_plugin=md_plugin)
    assert isinstance(pit, ParsedItem)
    assert pit.collection["b1"].data_type == "uint8"
    assert pit["b1"].driver_data == {"foo": "bar", "subdataset": "AA"}
    assert pit["b1"].subdataset == "AA"
    assert pit["b2"].subdataset == "AA"


def test_noassets_case(no_bands_stac) -> None:
    md = extract_collection_metadata(no_bands_stac)
    assert len(md.bands) == 0


def test_partial_proj_fallback(partial_proj_stac: pystac.item.Item) -> None:
    # Regression test for https://github.com/opendatacube/odc-stac/issues/251
    # Item declares proj extension but assets don't have proj:shape/transform
    item = partial_proj_stac
    assert has_proj_ext(item) is True

    # Extract collection metadata - should fall back to has_proj=False
    md = extract_collection_metadata(item)
    assert len(md.bands) > 0, "Should have found data bands with fallback"
    assert md.has_proj is False, "Should have fallen back to has_proj=False"

    # Parse item - should successfully extract bands
    parsed = parse_item(item, md)
    assert len(parsed.bands) == len(item.assets), "All assets should be parsed as bands"
    # Verify bands are parsed correctly
    for asset_name in item.assets:
        assert (asset_name, 1) in parsed.bands, f"Asset {asset_name} should be in bands"


def test_extract_md_raster_ext(
    sentinel_stac_ms_with_raster_ext: pystac.item.Item,
) -> None:
    item = sentinel_stac_ms_with_raster_ext

    md = extract_collection_metadata(item, STAC_CFG)

    assert md.aliases["red"] == [("B04", 1), ("visual", 1)]
    assert md.aliases["green"] == [("B03", 1), ("visual", 2)]
    assert md.aliases["blue"] == [("B02", 1), ("visual", 3)]


def test_parse_item(sentinel_stac_ms: pystac.item.Item) -> None:
    item0 = sentinel_stac_ms
    item = pystac.Item.from_dict(item0.to_dict())

    md = extract_collection_metadata(item, STAC_CFG)

    xx = parse_item(item, md)
    assert xx.datetime_range == (None, None)
    assert xx.datetime == item.datetime
    assert xx.nominal_datetime == item.datetime

    assert all(band in md for band in S2_ALL_BANDS)
    assert all((band, 1) in md for band in S2_ALL_BANDS)
    assert item not in xx

    assert set(n for n, _ in xx.bands) == S2_ALL_BANDS
    assert isinstance(xx["B02"], RasterSource)
    assert isinstance(xx["B01"], RasterSource)
    assert isinstance(xx["B05"], RasterSource)
    assert xx["B02"].geobox is not None
    assert xx["B02"] is xx[("B02", 1)]
    assert xx["B02"] is xx["B02.1"]
    assert xx.get("B02", None) is xx["B02.1"]

    assert xx.geoboxes() == xx.geoboxes(list(S2_ALL_BANDS))
    assert xx.geoboxes(["B02", "B03"]) == (xx["B02"].geobox,)
    assert xx.geoboxes(["B01", "B02", "B03"]) == (
        xx["B02"].geobox,
        xx["B01"].geobox,
    )
    assert xx.geoboxes() == (
        xx["B02"].geobox,  # 10m
        xx["B05"].geobox,  # 20m
        xx["B01"].geobox,  # 60m
    )

    (yy,) = list(parse_items(iter([item]), STAC_CFG))
    assert xx == yy

    # Test missing band case
    item = pystac.Item.from_dict(item0.to_dict())
    item.assets.pop("B01")
    xx = parse_item(item, md)
    assert "B01" not in xx
    assert xx.get("B01", None) is None


def test_parse_item_raster_ext(
    sentinel_stac_ms_with_raster_ext: pystac.item.Item,
) -> None:
    item = sentinel_stac_ms_with_raster_ext
    parsed = parse_item(item)
    assert isinstance(parsed[("visual", 2)], RasterSource)
    assert isinstance(parsed["visual.2"], RasterSource)
    assert isinstance(parsed[("visual", 3)], RasterSource)
    assert isinstance(parsed["visual.3"], RasterSource)
    assert parsed[("visual", 2)].band == 2
    assert parsed["visual.2"].band == 2
    assert parsed[("visual", 3)] is parsed["visual.3"]

    for (band, idx), b in parsed.bands.items():
        assert isinstance(b, RasterSource)
        assert idx == b.band
        assert band in S2_ALL_BANDS


@pytest.mark.xfail
def test_parse_no_absolute_href(relative_href_only: pystac.item.Item) -> None:
    # Currently pystac never returns `None` from attached asset
    # see: https://github.com/stac-utils/pystac/issues/754
    item = relative_href_only
    assert item.get_self_href() is None
    for asset in item.assets.values():
        assert pystac.utils.is_absolute_href(asset.href) is False
        assert asset.get_absolute_href() is None  # <<< asserts here, but shouldn't

    with pytest.raises(ValueError):
        _ = parse_item(item, extract_collection_metadata(item))


def test_parse_item_no_proj(sentinel_stac_ms: pystac.item.Item) -> None:
    item0 = sentinel_stac_ms
    item = pystac.Item.from_dict(item0.to_dict())
    item.stac_extensions.remove(ProjectionExtension.get_schema_uri())
    assert has_proj_ext(item) is False

    md = extract_collection_metadata(item, STAC_CFG)

    xx = parse_item(item, md)
    for band in xx.bands.values():
        assert isinstance(band, RasterSource)
        assert band.geobox is None

    assert xx.geoboxes() == ()

    assert _auto_load_params([xx] * 3) is None


def test_accessories_preserved(ga_landsat_stac: pystac.item.Item) -> None:
    item0 = ga_landsat_stac
    item = pystac.Item.from_dict(item0.to_dict())

    md = extract_collection_metadata(item, STAC_CFG)

    xx = parse_item(item, md)
    assert len(xx.accessories) == 3
    assert xx.accessories.get("thumbnail:nbart")
    assert xx.accessories.get("checksum:sha1")
    assert xx.accessories.get("metadata:processor")
    assert xx.strip().accessories == {}


@pytest.fixture
def parsed_item_s2(sentinel_stac_ms: pystac.item.Item):
    (item,) = parse_items([sentinel_stac_ms], STAC_CFG)
    yield item


def test_auto_load_params(parsed_item_s2: ParsedItem) -> None:
    xx = parsed_item_s2
    assert len(xx.geoboxes()) == 3
    crs = xx.geoboxes()[0].crs

    assert isinstance(xx["B01"], RasterSource)
    assert isinstance(xx["B02"], RasterSource)
    assert isinstance(xx["B05"], RasterSource)

    _gbox_10m = xx["B02"].geobox
    _gbox_20m = xx["B05"].geobox
    _gbox_60m = xx["B01"].geobox
    assert _gbox_10m is not None
    assert _gbox_20m is not None
    assert _gbox_60m is not None
    _10m = _gbox_10m.resolution
    _20m = _gbox_20m.resolution
    _60m = _gbox_60m.resolution
    _edge = AnchorEnum.EDGE

    assert _10m.xy == (10, -10)
    assert _20m.xy == (20, -20)
    assert _60m.xy == (60, -60)

    assert _auto_load_params([]) is None
    assert _auto_load_params([xx]) == (crs, _10m, _edge, _gbox_10m)
    assert _auto_load_params([xx] * 3) == (crs, _10m, _edge, _gbox_10m)

    assert _auto_load_params([xx], ["B01"]) == (crs, _60m, _edge, _gbox_60m)
    assert _auto_load_params([xx] * 3, ["B01", "B05", "B06"]) == (
        crs,
        _20m,
        _edge,
        _gbox_20m,
    )
    assert _auto_load_params([xx] * 3, ["B01", "B04"]) == (crs, _10m, _edge, _gbox_10m)


def test_norm_geom(gpd_iso3) -> None:
    g = geom.box(0, -1, 10, 1, "epsg:4326")

    assert _normalize_geometry(g) is g
    assert _normalize_geometry(g.geom) == g
    assert _normalize_geometry(g.json) == g

    assert _normalize_geometry(g.geojson()) == g
    assert (
        _normalize_geometry(dict(type="FeatureCollection", features=[g.geojson()])) == g
    )

    g = gpd_iso3("AUS")
    assert g.crs == "epsg:4326"
    assert _normalize_geometry(g).crs == "epsg:4326"

    g = gpd_iso3("AUS", "epsg:3577")
    assert g.crs == "epsg:3577"
    assert _normalize_geometry(g).crs == "epsg:3577"

    with pytest.raises(ValueError):
        _ = _normalize_geometry({})  # not a valid geojson

    with pytest.raises(ValueError):
        _ = _normalize_geometry(object())  # Can't interpret value as geometry


def test_output_geobox(gpd_iso3, parsed_item_s2: ParsedItem) -> None:
    au = gpd_iso3("AUS", "epsg:3577")

    gbox = output_geobox([], geopolygon=au, resolution=100, crs="epsg:3857")
    assert gbox is not None
    assert gbox.crs == "epsg:3857"
    assert gbox.resolution.xy == (100, -100)

    # default CRS to that of the polygon
    gbox = output_geobox([], geopolygon=au, resolution=100)
    assert gbox is not None
    assert gbox.crs == "epsg:3577"
    assert gbox.resolution.xy == (100, -100)

    gbox = output_geobox([parsed_item_s2], geopolygon=au, resolution=100)
    assert gbox is not None
    assert gbox.crs == parsed_item_s2.crs()
    assert gbox.resolution.xy == (100, -100)

    gbox = output_geobox([parsed_item_s2])
    assert gbox is not None
    assert gbox.crs == parsed_item_s2.crs()
    assert gbox.resolution.xy == (10, -10)

    # like/gbox
    assert output_geobox([], geobox=gbox) == gbox
    assert output_geobox([], like=gbox) == gbox
    assert output_geobox([], like=xr_zeros(gbox[:10, :20])) == gbox[:10, :20]

    # no resolution/crs
    assert output_geobox([], bbox=(0, 1, 2, 3)) is None
    assert output_geobox([], bbox=(0, 1, 2, 3), resolution=10) is None
    assert output_geobox([], bbox=(0, 1, 2, 3), crs="epsg:4326") is None

    # lon-lat/x-y/bbox
    gbox = GeoBox.from_bbox((0, -10, 100, 25), resolution=1)
    assert gbox.boundingbox == (0, -10, 100, 25)
    bbox = gbox.boundingbox

    assert gbox == output_geobox(
        [],
        x=bbox.range_x,
        y=bbox.range_y,
        resolution=gbox.resolution,
        crs=gbox.crs,
    )

    assert gbox == output_geobox(
        [],
        lon=bbox.range_x,
        lat=bbox.range_y,
        resolution=gbox.resolution,
        crs=gbox.crs,
    )

    assert gbox == output_geobox(
        [],
        bbox=bbox.bbox,
        resolution=gbox.resolution,
        crs=gbox.crs,
    )

    assert gbox == output_geobox(
        [],
        bbox=bbox.bbox,
        resolution=gbox.resolution,
        crs=gbox.crs,
        align=0,
    )


def test_output_geobox_from_items() -> None:
    cc = 0

    def mk_item(gbox: GeoBox, time="2020-01-10"):
        nonlocal cc
        cc = cc + 1
        return mk_parsed_item(
            [b_("b1", geobox=gbox), b_("b2", geobox=gbox)], time, id=f"item-{cc}"
        )

    gboxes = [GBOX, GBOX.left, GBOX.right.pad(3)]

    gbox = output_geobox([mk_item(gbox) for gbox in gboxes])
    assert gbox is not None
    assert gbox.crs == GBOX.crs
    assert geobox_union_conservative(gboxes) == gbox


@pytest.mark.parametrize(
    "kw",
    [
        # not enough args
        {"x": (0, 10)},
        {"y": (0, 10)},
        {"lon": (0, 1)},
        {"lat": (0, 1)},
        {"x": (0, 1), "y": (1, 2)},
        # too many args
        {"lat": (0, 1), "lon": (1, 2), "x": (3, 4), "y": (5, 6)},
        {"lat": (0, 1), "lon": (1, 2), "bbox": (0, 1, 2, 3)},
        {"lat": (0, 1), "lon": (1, 2), "geopolygon": geom.box(0, 0, 1, 1, "epsg:4326")},
        {"bbox": (0, 0, 1, 1), "geopolygon": geom.box(0, 0, 1, 1, "epsg:4326")},
        # bad args
        {"like": object()},
    ],
)
def test_output_gbox_bads(kw) -> None:
    with pytest.raises(ValueError):
        _ = output_geobox([], **kw)


def test_mk_parsed_item() -> None:
    fmt = "%Y-%m-%d"
    item = mk_parsed_item(
        [b_("b1"), b_("b2")],
        "2020-01-10",
        start_datetime="2020-01-01",
        end_datetime="2020-01-31",
    )

    assert item.datetime is not None
    assert item.datetime.strftime(fmt) == "2020-01-10"
    assert item.datetime_range[0] is not None
    assert item.datetime_range[0].strftime(fmt) == "2020-01-01"
    assert item.datetime_range[1] is not None
    assert item.datetime_range[1].strftime(fmt) == "2020-01-31"
    assert item.geometry is None
    assert item.crs() is None
    assert item.collection.has_proj is False

    assert set(item.bands) == set([("b1", 1), ("b2", 1)])
    assert item["b1"].uri.endswith("b1.tif")

    item = mk_parsed_item(
        [b_("b1"), b_("b2")],
        datetime=None,
        start_datetime="2020-01-01",
        end_datetime="2020-01-31",
    )

    assert item.datetime is None
    assert item.datetime_range[0] is not None
    assert item.datetime_range[0].strftime(fmt) == "2020-01-01"
    assert item.datetime_range[1] is not None
    assert item.datetime_range[1].strftime(fmt) == "2020-01-31"

    item = mk_parsed_item(
        [b_("b1"), b_("b2")],
        "2020-01-10",
        start_datetime="2020-01-01",
        end_datetime=None,
    )
    assert item.datetime is not None
    assert item.datetime.strftime(fmt) == "2020-01-10"
    assert item.datetime_range[0] is not None
    assert item.datetime_range[0].strftime(fmt) == "2020-01-01"
    assert item.datetime_range[1] is None

    gbox = GeoBox.from_bbox((-20, -10, 20, 10), "epsg:3857", shape=(200, 400))
    item = mk_parsed_item(
        [b_("b1", geobox=gbox), b_("b2", geobox=gbox)],
        "2020-01-10",
    )
    assert item.geometry is not None
    assert item.geometry.crs == "epsg:4326"
    assert item.crs() == "epsg:3857"
    assert item.geoboxes() == (gbox,)
    assert item.collection.has_proj is True


@pytest.mark.parametrize(
    "parsed_item",
    [
        mk_parsed_item(
            [b_("band")],
            None,
            start_datetime="2020-01-01",
            end_datetime="2021-12-31T23:59:59.9999999Z",
        ),
        mk_parsed_item([b_("b1"), b_("b2", nodata=10)], "2020-01-01"),
        mk_parsed_item(
            [
                b_("b1", dtype="float32", geobox=GBOX),
                b_("b2", nodata=10, geobox=GBOX),
            ],
            "2020-01-01",
        ),
        mk_parsed_item(
            [
                b_("b1", dtype="float32", geobox=GBOX),
                b_("b2", dtype="int32", nodata=-99, geobox=GBOX.zoom_out(2)),
            ],
            "2020-01-01",
            start_datetime="2020-01-01",
            end_datetime="2021-12-31T23:59:59.9999999Z",
            href="file:///date/item/1.json",
        ),
    ],
)
def test_round_trip(parsed_item: ParsedItem) -> None:
    item = to_stac_item(parsed_item)
    md = extract_collection_metadata(item)

    assert parsed_item.collection == md
    assert parsed_item == parse_item(item, md)


def test_usgs_v1_1_1_aliases(usgs_landsat_stac_v1_1_1: pystac.Item) -> None:
    parsed_item = next(parse_items([usgs_landsat_stac_v1_1_1]))
    collection = parsed_item.collection
    assert collection.aliases == {
        "B1": [("blue", 1)],
        "B2": [("green", 1)],
        "B3": [("red", 1)],
        "B4": [("nir08", 1)],
        "B5": [("swir16", 1)],
        "B7": [("swir22", 1)],
    }


def test_most_common_gbox() -> None:
    gbox = GeoBox.from_bbox((0, 0, 100, 200), resolution=10, crs=3857)
    assert _most_common_gbox(
        [gbox, gbox.center_pixel, gbox[:1, :1], gbox.zoom_out(1.3)]
    ) == (
        gbox.crs,
        gbox.resolution,
        AnchorEnum.EDGE,
        None,
    )
    # not enough consensus for anchor
    # fallback to EDGE aligned
    assert _most_common_gbox(
        [
            gbox.translate_pix(-0.3, -0.2),
            gbox.center_pixel.translate_pix(-0.1, -0.4),
            gbox.zoom_out(1.3),
            gbox.to_crs(4326),
        ],
        1 / 4 + 0.1,
    ) == (
        gbox.crs,
        gbox.resolution,
        AnchorEnum.EDGE,
        None,
    )

    # CENTER
    gbox = GeoBox.from_bbox(
        (0, 0, 100, 200), resolution=10, crs=3857, anchor=AnchorEnum.CENTER
    )
    assert _most_common_gbox(
        [gbox, gbox.center_pixel, gbox[:1, :1], gbox.zoom_out(1.3)]
    ) == (
        gbox.crs,
        gbox.resolution,
        AnchorEnum.CENTER,
        None,
    )
